import dataclasses
import inspect
import itertools
import logging
import typing as t
from collections import defaultdict

import yaml

from flama import routing, schemas, types
from flama.schemas import Schema, openapi
from flama.schemas.data_structures import Parameter
from flama.url import RegexPath

logger = logging.getLogger(__name__)

__all__ = ["SchemaRegistry", "SchemaGenerator"]


@dataclasses.dataclass(frozen=True)
class EndpointInfo:
    path: str
    method: str
    func: t.Callable = dataclasses.field(repr=False)
    query_parameters: dict[str, Parameter] = dataclasses.field(repr=False)
    path_parameters: dict[str, Parameter] = dataclasses.field(repr=False)
    body_parameter: t.Optional[Parameter] = dataclasses.field(repr=False)
    response_parameter: Parameter = dataclasses.field(repr=False)


@dataclasses.dataclass(frozen=True)
class SchemaInfo:
    name: str
    schema: t.Any

    @property
    def ref(self) -> str:
        return f"#/components/schemas/{self.name}"

    def json_schema(self, names: dict[int, str]) -> types.JSONSchema:
        return Schema(self.schema).json_schema(names)


class SchemaRegistry(dict[int, SchemaInfo]):
    def __init__(self, schemas: t.Optional[dict[str, schemas.Schema]] = None):
        super().__init__()

        for name, schema in (schemas or {}).items():
            self.register(schema, name)

    def __contains__(self, item: t.Any) -> bool:
        return super().__contains__(id(schemas.Schema(item).unique_schema))

    def __getitem__(self, item: t.Any) -> SchemaInfo:
        """Lookup method that allows using Schema classes or instances.

        :param item: Schema to look for.
        :return: Registered schema.
        """
        return super().__getitem__(id(schemas.Schema(item).unique_schema))

    @property
    def names(self) -> dict[int, str]:
        """Returns a dictionary mapping schema ids to their names.

        :return: Schema names.
        """
        return {k: v.name for k, v in self.items()}

    @t.no_type_check
    def _get_schema_references_from_schema(self, schema: t.Union[openapi.Schema, openapi.Reference]) -> list[str]:
        if isinstance(schema, openapi.Reference):
            return [schema.ref]

        result = []

        if "$ref" in schema:
            result.append(schema["$ref"])

        if schema.get("type", "") == "array":
            items = schema.get("items", {})
            if isinstance(items, dict) and "$ref" in items:
                result.append(items["$ref"])

        for composer in ("allOf", "anyOf", "oneOf"):
            composer_schemas = schema.get(composer, [])
            if isinstance(composer_schemas, list):
                result += [
                    ref
                    for x in composer_schemas
                    if x and isinstance(x, dict)
                    for ref in self._get_schema_references_from_schema(openapi.Schema(x))
                ]

        props = schema.get("properties", {})
        if isinstance(props, dict):
            result += [
                ref
                for x in props.values()
                if x and isinstance(x, dict)
                for ref in self._get_schema_references_from_schema(openapi.Schema(x))
            ]

        return result

    def _get_schema_references_from_path(self, path: openapi.Path) -> list[str]:
        return [y for x in path.operations.values() for y in self._get_schema_references_from_operation(x)]

    def _get_schema_references_from_operation(self, operation: openapi.Operation) -> list[str]:
        return [
            *self._get_schema_references_from_operation_parameters(operation.parameters),
            *self._get_schema_references_from_operation_request_body(operation.requestBody),
            *self._get_schema_references_from_operation_callbacks(operation.callbacks),
            *self._get_schema_references_from_operation_responses(operation.responses),
        ]

    def _get_schema_references_from_operation_responses(self, responses: openapi.Responses) -> list[str]:
        refs = []

        for response in [x for x in responses.values() if x.content]:
            for media_type in [
                x
                for x in response.content.values()  # type: ignore[union-attr]
                if isinstance(x, openapi.MediaType) and x.schema
            ]:
                refs += self._get_schema_references_from_schema(media_type.schema)  # type: ignore[arg-type]

        return refs

    def _get_schema_references_from_operation_callbacks(
        self, callbacks: t.Optional[dict[str, t.Union[openapi.Callback, openapi.Reference]]]
    ) -> list[str]:
        refs = []

        if callbacks:
            for callback in callbacks.values():
                if isinstance(callback, openapi.Reference):
                    refs.append(callback.ref)
                else:
                    for callback_path in callback.values():
                        refs += self._get_schema_references_from_path(callback_path)

        return refs

    def _get_schema_references_from_operation_request_body(
        self, request_body: t.Optional[t.Union[openapi.RequestBody, openapi.Reference]]
    ) -> list[str]:
        refs = []

        if request_body:
            if isinstance(request_body, openapi.Reference):
                refs.append(request_body.ref)
            else:
                for media_type in [x for x in request_body.content.values() if isinstance(x, openapi.MediaType)]:
                    refs += self._get_schema_references_from_schema(media_type.schema)  # type: ignore[arg-type]

        return refs

    def _get_schema_references_from_operation_parameters(
        self, parameters: t.Optional[list[t.Union[openapi.Parameter, openapi.Reference]]]
    ) -> list[str]:
        refs = []

        if parameters:
            for param in parameters:
                if isinstance(param, openapi.Reference):
                    refs.append(param.ref)
                else:
                    refs += self._get_schema_references_from_schema(param.schema)  # type: ignore[arg-type]

        return refs

    def used(self, spec: openapi.OpenAPISpec) -> dict[int, SchemaInfo]:
        """
        Generate a dict containing used schemas.

        :param spec: API Schema.
        :return: Used schemas.
        """
        refs_from_spec = {
            x.split("/")[-1] for path in spec.spec.paths.values() for x in self._get_schema_references_from_path(path)
        }
        used_schemas = {k: v for k, v in self.items() if v.name in refs_from_spec}
        refs_from_schemas = {
            x.split("/")[-1]
            for schema in used_schemas.values()
            for x in self._get_schema_references_from_schema(openapi.Schema(schema.json_schema(self.names)))
        }
        used_schemas.update({k: v for k, v in self.items() if v.name in refs_from_schemas})

        for child_schema in [y for x in used_schemas.values() for y in schemas.Schema(x.schema).nested_schemas()]:
            schema = schemas.Schema(child_schema)
            instance = schema.unique_schema

            if instance not in used_schemas:
                used_schemas[id(instance)] = self[instance]

        return used_schemas

    def register(self, schema: schemas.Schema, name: t.Optional[str] = None) -> int:
        """
        Register a new Schema to this registry.

        :param schema: Schema object or class.
        :param name: Schema name.
        :return: Schema ID.
        """
        if schema in self:
            raise ValueError("Schema is already registered.")

        s = schemas.Schema(schema)

        try:
            schema_name = name or s.name
        except ValueError as e:  # pragma: no cover
            raise ValueError("Cannot infer schema name.") from e

        schema_instance = s.unique_schema
        schema_id = id(schema_instance)
        self[schema_id] = SchemaInfo(name=schema_name, schema=schema_instance)

        for child_schema in [schemas.Schema(x) for x in s.nested_schemas() if x not in self]:
            self.register(schema=child_schema.schema, name=child_schema.name)

        return schema_id

    def get_openapi_ref(
        self, element: schemas.Schema, multiple: t.Optional[bool] = None
    ) -> t.Union[openapi.Schema, openapi.Reference]:
        """
        Builds the reference for a single schema or the array schema containing the reference.

        :param element: Schema to use as reference.
        :param multiple: True for building a schema containing an array of references instead of a single reference.
        :return: Reference or array schema.
        """
        reference = self[element].ref

        if multiple is True:
            return openapi.Schema({"items": {"$ref": reference}, "type": "array"})
        else:
            return openapi.Reference(ref=reference)


class SchemaGenerator:
    def __init__(
        self,
        title: str,
        version: str,
        description: t.Optional[str] = None,
        terms_of_service: t.Optional[str] = None,
        contact_name: t.Optional[str] = None,
        contact_url: t.Optional[str] = None,
        contact_email: t.Optional[str] = None,
        license_name: t.Optional[str] = None,
        license_url: t.Optional[str] = None,
        schemas: t.Optional[dict] = None,
    ):
        contact = (
            openapi.Contact(name=contact_name, url=contact_url, email=contact_email)
            if contact_name or contact_url or contact_email
            else None
        )

        license = openapi.License(name=license_name, url=license_url) if license_name else None

        self.spec = openapi.OpenAPISpec(
            title=title,
            version=version,
            description=description,
            terms_of_service=terms_of_service,
            contact=contact,
            license=license,
        )

        # Builtin definitions
        self.schemas = SchemaRegistry(schemas=schemas)

    def get_endpoints(  # type: ignore[override]
        self, routes: list[routing.BaseRoute], base_path: str = ""
    ) -> dict[str, list[EndpointInfo]]:
        """
        Given the routes, yields the following information:

        - path
            eg: /users/
        - http_method
            one of 'get', 'post', 'put', 'patch', 'delete', 'options'
        - func
            method ready to extract the docstring

        :param routes: A list of routes.
        :param base_path: The base endpoints path.
        :return: Data structure that contains metadata from every route.
        """
        endpoints_info: dict[str, list[EndpointInfo]] = defaultdict(list)

        for route in routes:
            path = RegexPath(base_path + route.path.path).template

            if isinstance(route, routing.Route) and route.include_in_schema:
                if inspect.isfunction(route.endpoint) or inspect.ismethod(route.endpoint):
                    for method in route.methods or ["GET"]:
                        if method == "HEAD":
                            continue

                        endpoints_info[path].append(
                            EndpointInfo(
                                path=path,
                                method=method.lower(),
                                func=route.endpoint,
                                query_parameters=route.parameters.query.get(method, {}),
                                path_parameters=route.parameters.path.get(method, {}),
                                body_parameter=route.parameters.body.get(method),
                                response_parameter=route.parameters.response[method],
                            )
                        )
                else:
                    for method in [x.lower() for x in route.methods]:
                        if not hasattr(route.endpoint, method):
                            continue

                        func = getattr(route.endpoint, method)
                        endpoints_info[path].append(
                            EndpointInfo(
                                path=path,
                                method=method.lower(),
                                func=func,
                                query_parameters=route.parameters.query.get(method.upper(), {}),
                                path_parameters=route.parameters.path.get(method.upper(), {}),
                                body_parameter=route.parameters.body.get(method.upper()),
                                response_parameter=route.parameters.response[method.upper()],
                            )
                        )
            elif isinstance(route, routing.Mount):
                endpoints_info.update(self.get_endpoints(route.routes, base_path=path))

        return endpoints_info

    def _build_endpoint_parameters(
        self, endpoint: EndpointInfo, metadata: dict[str, t.Any]
    ) -> t.Optional[list[openapi.Parameter]]:
        if not endpoint.query_parameters and not endpoint.path_parameters:
            return None

        return [
            openapi.Parameter(
                schema=openapi.Schema(field.field.json_schema),
                name=field.name,
                in_=field.location.name,
                required=field.required,
                **{
                    x: y.get(x)
                    for y in [x for x in metadata.get("parameters", []) if x.get("name") == field.name]
                    if y
                    for x in (
                        "description",
                        "deprecated",
                        "allowEmptyValue",
                        "style",
                        "explode",
                        "allowReserved",
                        "example",
                    )
                },
            )
            for field in itertools.chain(endpoint.query_parameters.values(), endpoint.path_parameters.values())
        ]

    def _build_endpoint_body(
        self, endpoint: EndpointInfo, metadata: dict[str, t.Any]
    ) -> t.Optional[openapi.RequestBody]:
        content = {k: v for k, v in metadata.get("requestBody", {}).get("content", {}).items()}

        if endpoint.body_parameter:
            if endpoint.body_parameter.schema.schema not in self.schemas:
                self.schemas.register(schema=endpoint.body_parameter.schema.schema)

            content["application/json"] = openapi.MediaType(
                schema=self.schemas.get_openapi_ref(
                    endpoint.body_parameter.schema.schema, multiple=endpoint.body_parameter.multiple
                ),
            )

        if not content:
            return None

        return openapi.RequestBody(
            content=content,
            **{x: metadata.get("requestBody", {}).get(x) for x in ("description", "required")},
        )

    def _build_endpoint_responses(self, endpoint: EndpointInfo, metadata: dict[str, t.Any]) -> openapi.Responses:
        responses = metadata.get("responses", {})
        try:
            main_response_code = next(iter(responses.keys()))
            assert 100 <= int(main_response_code) < 600
        except (ValueError, AssertionError, StopIteration):
            main_response_code = 200
            responses[main_response_code] = {
                "description": "Description not provided.",
            }
            logger.warning(
                'OpenAPI description not provided in docstring for main response in endpoint "%s", adding a standard '
                '"%s" response',
                main_response_code,
                endpoint.path,
            )

        if endpoint.response_parameter.schema.schema:
            if endpoint.response_parameter.schema.schema not in self.schemas:
                self.schemas.register(schema=endpoint.response_parameter.schema.schema)

            responses[main_response_code]["content"] = {
                **responses[main_response_code].get("content", {}),
                "application/json": {
                    "schema": self.schemas.get_openapi_ref(
                        endpoint.response_parameter.schema.schema, multiple=endpoint.response_parameter.multiple
                    )
                },
            }

        responses["default"] = {
            "description": "Unexpected error.",
            **responses.get("default", {}),
            "content": {
                "application/json": {"schema": self.schemas.get_openapi_ref(schemas.schemas.APIError, multiple=False)}
            },
        }

        return openapi.Responses(
            {
                str(code): openapi.Response(
                    description=response["description"],
                    headers=response.get("headers"),
                    content={
                        mime: openapi.MediaType(
                            schema=media_type.get("schema"),
                            example=media_type.get("example"),
                            examples=media_type.get("examples"),
                            encoding=media_type.get("encoding"),
                        )
                        for mime, media_type in response.get("content", {}).items()
                    },
                    links=response.get("links"),
                )
                for code, response in responses.items()
            }
        )

    def _parse_docstring(self, func: t.Callable) -> dict[t.Any, t.Any]:
        """Given a function, parse the docstring as YAML and return a dictionary of info.

        :param func: Function to analyze docstring.
        :return: Schema extracted.
        """
        try:
            # It's possible to define a standard docstring along with the schema definition, for doing so the schema
            # should start with a line with three dashes "---" as it's the usual notation for starting a yaml file.
            schema = yaml.safe_load(func.__doc__.split("---")[-1])  # type: ignore[union-attr]
        except AttributeError:
            schema = None

        if not isinstance(schema, dict):
            raise ValueError

        return schema

    def get_operation_schema(self, endpoint: EndpointInfo) -> openapi.Operation:
        try:
            docstring_info = self._parse_docstring(endpoint.func)
        except ValueError:
            docstring_info = {}

        # Query and Path parameters
        parameters = self._build_endpoint_parameters(endpoint, docstring_info)

        # Body
        request_body = self._build_endpoint_body(endpoint, docstring_info)

        # Response
        responses = self._build_endpoint_responses(endpoint, docstring_info)

        return openapi.Operation(
            responses=responses,
            parameters=parameters,  # type: ignore[arg-type]
            requestBody=request_body,
            **{
                x: docstring_info.get(x)
                for x in ("tags", "summary", "description", "externalDocs", "operationId", "deprecated")
            },
        )

    def get_api_schema(self, routes: list[routing.BaseRoute]) -> dict[str, t.Any]:
        endpoints_info = self.get_endpoints(routes)

        for path, endpoints in endpoints_info.items():
            operations = {}
            for endpoint in endpoints:
                try:
                    operations[endpoint.method] = self.get_operation_schema(endpoint)
                except Exception:
                    logger.error("Cannot generate schema for endpoint %s", endpoint)

            if operations:
                self.spec.add_path(path, openapi.Path(**operations))  # type: ignore[arg-type]

        for schema in self.schemas.used(self.spec).values():
            self.spec.add_schema(schema.name, openapi.Schema(schema.json_schema(self.schemas.names)))

        api_schema: dict[str, t.Any] = self.spec.asdict()

        return api_schema



import typing as t

from flama import exceptions
from flama.codecs.base import WebsocketsCodec

if t.TYPE_CHECKING:
    from flama import types

__all__ = ["TextCodec"]


class TextCodec(WebsocketsCodec):
    encoding = "text"

    async def decode(self, item: "types.Message", **options):
        if "text" not in item:
            raise exceptions.DecodeError("Expected text websocket messages")

        return item["text"]



import typesystem
from typesystem import Schema
from typesystem.fields import Field

from flama.schemas._libs.typesystem import fields, schemas
from flama.schemas._libs.typesystem.adapter import TypesystemAdapter

lib = typesystem
adapter = TypesystemAdapter()

__all__ = ["Field", "Schema", "fields", "adapter", "lib", "schemas"]



import abc
import asyncio
import inspect
import typing as t


class PaginationDecoratorFactory:
    PARAMETERS: list[inspect.Parameter]

    @classmethod
    def decorate(cls, func: t.Callable, schema: t.Any) -> t.Callable:
        func_signature = inspect.signature(func)
        if "kwargs" not in func_signature.parameters:
            raise TypeError("Paginated views must define **kwargs param")

        decorated_func = (
            cls._decorate_async(func, schema) if asyncio.iscoroutinefunction(func) else cls._decorate_sync(func, schema)
        )

        decorated_func.__signature__ = inspect.Signature(  # type: ignore
            parameters=[
                *[v for k, v in func_signature.parameters.items() if k != "kwargs"],
                *cls.PARAMETERS,
            ],
            return_annotation=schema,
        )

        return decorated_func

    @classmethod
    @abc.abstractmethod
    def _decorate_async(cls, func: t.Callable, schema: t.Any) -> t.Callable:
        ...

    @classmethod
    @abc.abstractmethod
    def _decorate_sync(cls, func: t.Callable, schema: t.Any) -> t.Callable:
        ...



import asyncio
import inspect
from functools import wraps

from flama import exceptions, schemas

__all__ = ["output_validation"]


def output_validation(error_cls=exceptions.ValidationError, error_status_code=500):
    """
    Validate view output using schema annotated as function's return.

    :param error_cls: Error class to be raised when validation fails. Errors dict will be passed through 'detail' param.
    :param error_status_code: HTTP status code assigned to response when it fails to validate output, default 500.
    :raises exceptions.ValidationError: if output validation fails.
    """

    def outer_decorator(func):
        try:
            schema = schemas.Schema.from_type(inspect.signature(func).return_annotation)
        except Exception as e:
            raise TypeError(f"Invalid return signature for function '{func}'") from e

        @wraps(func)
        async def inner_decorator(*args, **kwargs):
            response = await func(*args, **kwargs) if asyncio.iscoroutinefunction(func) else func(*args, **kwargs)

            try:
                # Use output schema to validate the data
                schema.validate(schema.dump(response))
            except schemas.SchemaValidationError as e:
                raise error_cls(detail=e.errors, status_code=error_status_code)
            except Exception as e:
                raise error_cls(
                    detail=f"Error serializing response before validation: {str(e)}", status_code=error_status_code
                )

            return response

        return inner_decorator

    return outer_decorator



import abc
import typing as t

if t.TYPE_CHECKING:
    from flama.serialize.data_structures import Artifacts, Metadata

__all__ = ["Model"]


class Model:
    def __init__(self, model: t.Any, meta: "Metadata", artifacts: t.Optional["Artifacts"]):
        self.model = model
        self.meta = meta
        self.artifacts = artifacts

    def inspect(self) -> t.Any:
        return {"meta": self.meta.to_dict(), "artifacts": self.artifacts}

    @abc.abstractmethod
    def predict(self, x: t.Any) -> t.Any:
        ...



from flama.types.applications import *  # noqa
from flama.types.asgi import *  # noqa
from flama.types.endpoints import *  # noqa
from flama.types.http import *  # noqa
from flama.types.json import *  # noqa
from flama.types.websockets import *  # noqa



import typing as t

from flama import exceptions
from flama.models.base import Model

try:
    import numpy as np  # type: ignore
except Exception:  # pragma: no cover
    np = None

try:
    import tensorflow as tf  # type: ignore
except Exception:  # pragma: no cover
    tf = None


class TensorFlowModel(Model):
    def predict(self, x: list[list[t.Any]]) -> t.Any:
        if np is None:  # noqa
            raise exceptions.FrameworkNotInstalled("numpy")

        if tf is None:  # noqa
            raise exceptions.FrameworkNotInstalled("tensorflow")

        try:
            return self.model.predict(np.array(x)).tolist()
        except (tf.errors.OpError, ValueError):  # type: ignore
            raise exceptions.HTTPException(status_code=400)



import pydantic
from pydantic import BaseModel as Schema
from pydantic import Field

from flama.schemas._libs.pydantic import fields, schemas
from flama.schemas._libs.pydantic.adapter import PydanticAdapter

lib = pydantic
adapter = PydanticAdapter()

__all__ = ["Field", "Schema", "fields", "adapter", "lib", "schemas"]



import enum

from flama import compat


class Framework(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
    """ML formats available for Flama serialization."""

    sklearn = enum.auto()
    tensorflow = enum.auto()
    torch = enum.auto()
    keras = enum.auto()



import typing as t
from http import HTTPStatus

from flama import exceptions, http, schemas
from flama.ddd import exceptions as ddd_exceptions
from flama.resources import data_structures
from flama.resources.rest import RESTResource, RESTResourceType
from flama.resources.routing import resource_method
from flama.resources.workers import FlamaWorker

__all__ = ["CreateMixin", "RetrieveMixin", "UpdateMixin", "DeleteMixin", "ListMixin", "DropMixin", "CRUDResourceType"]


class CreateMixin:
    @classmethod
    def _add_create(
        cls,
        name: str,
        verbose_name: str,
        rest_schemas: data_structures.Schemas,
        rest_model: data_structures.Model,
        **kwargs,
    ) -> dict[str, t.Any]:
        @resource_method("/", methods=["POST"], name="create")
        async def create(
            self,
            worker: FlamaWorker,
            resource: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.input.schema)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.output.schema)]:
            if resource.get(rest_model.primary_key.name) is None:
                resource.pop(rest_model.primary_key.name, None)

            async with worker:
                repository = worker.repositories[self._meta.name]
                try:
                    result = await repository.create(resource)
                except ddd_exceptions.IntegrityError as e:
                    raise exceptions.HTTPException(status_code=HTTPStatus.BAD_REQUEST, detail=str(e))

            return http.APIResponse(  # type: ignore[return-value]
                schema=rest_schemas.output.schema, content=result[0], status_code=HTTPStatus.CREATED
            )

        create.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Create a new resource
            description:
                Create a new resource in this collection.
            responses:
                201:
                    description:
                        Resource created successfully.
                400:
                    description:
                        Resource already exists or cannot be created.
        """

        return {"_create": create}


class RetrieveMixin:
    @classmethod
    def _add_retrieve(
        cls,
        name: str,
        verbose_name: str,
        rest_schemas: data_structures.Schemas,
        rest_model: data_structures.Model,
        **kwargs,
    ) -> dict[str, t.Any]:
        @resource_method("/{resource_id}/", methods=["GET"], name="retrieve")
        async def retrieve(
            self,
            worker: FlamaWorker,
            resource_id: rest_model.primary_key.type,  # type: ignore
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.output.schema)]:  # type: ignore
            try:
                async with worker:
                    repository = worker.repositories[self._meta.name]
                    return await repository.retrieve(**{rest_model.primary_key.name: resource_id})
            except ddd_exceptions.NotFoundError as e:
                raise exceptions.HTTPException(status_code=HTTPStatus.NOT_FOUND, detail=str(e))

        retrieve.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Retrieve a resource
            description:
                Retrieve a resource from this collection.
            responses:
                200:
                    description:
                        Resource found and retrieved.
                404:
                    description:
                        Resource not found.
        """

        return {"_retrieve": retrieve}


class UpdateMixin:
    @classmethod
    def _add_update(
        cls,
        name: str,
        verbose_name: str,
        rest_schemas: data_structures.Schemas,
        rest_model: data_structures.Model,
        **kwargs,
    ) -> dict[str, t.Any]:
        @resource_method("/{resource_id}/", methods=["PUT"], name="update")
        async def update(
            self,
            worker: FlamaWorker,
            resource_id: rest_model.primary_key.type,  # type: ignore
            resource: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.input.schema)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.output.schema)]:
            resource[rest_model.primary_key.name] = resource_id
            async with worker:
                try:
                    repository = worker.repositories[self._meta.name]
                    await repository.delete(**{rest_model.primary_key.name: resource_id})
                except ddd_exceptions.NotFoundError as e:
                    raise exceptions.HTTPException(status_code=HTTPStatus.NOT_FOUND, detail=str(e))

                try:
                    result = await repository.create(resource)
                except ddd_exceptions.IntegrityError as e:
                    raise exceptions.HTTPException(status_code=HTTPStatus.BAD_REQUEST, detail=str(e))

            return result[0]

        update.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Update a resource
            description:
                Update a resource in this collection.
            responses:
                200:
                    description:
                        Resource updated successfully.
                400:
                    description:
                        Wrong input data.
                404:
                    description:
                        Resource not found.
        """

        return {"_update": update}


class PartialUpdateMixin:
    @classmethod
    def _add_partial_update(
        cls,
        name: str,
        verbose_name: str,
        rest_schemas: data_structures.Schemas,
        rest_model: data_structures.Model,
        **kwargs,
    ) -> dict[str, t.Any]:
        @resource_method("/{resource_id}/", methods=["PATCH"], name="partial-update")
        async def partial_update(
            self,
            worker: FlamaWorker,
            resource_id: rest_model.primary_key.type,  # type: ignore
            resource: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.input.schema, partial=True)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.output.schema)]:
            resource[rest_model.primary_key.name] = resource_id
            async with worker:
                repository = worker.repositories[self._meta.name]
                try:
                    result = await repository.update(resource, **{rest_model.primary_key.name: resource_id})
                except ddd_exceptions.IntegrityError as e:
                    raise exceptions.HTTPException(status_code=HTTPStatus.BAD_REQUEST, detail=str(e))

                if not result:
                    raise exceptions.HTTPException(status_code=HTTPStatus.NOT_FOUND)

            return result[0]

        partial_update.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Partially update a resource
            description:
                Partially update a resource in this collection. Only the specified fields will be replaced, keeping the
                 rest, so no one is required.
            responses:
                200:
                    description:
                        Resource updated successfully.
                400:
                    description:
                        Wrong input data.
                404:
                    description:
                        Resource not found.
        """

        return {"_partial_update": partial_update}


class DeleteMixin:
    @classmethod
    def _add_delete(cls, name: str, verbose_name: str, rest_model: data_structures.Model, **kwargs) -> dict[str, t.Any]:
        @resource_method("/{resource_id}/", methods=["DELETE"], name="delete")
        async def delete(self, worker: FlamaWorker, resource_id: rest_model.primary_key.type):  # type: ignore
            try:
                async with worker:
                    repository = worker.repositories[self._meta.name]
                    await repository.delete(**{rest_model.primary_key.name: resource_id})
            except ddd_exceptions.NotFoundError as e:
                raise exceptions.HTTPException(status_code=HTTPStatus.NOT_FOUND, detail=str(e))

            return http.APIResponse(status_code=HTTPStatus.NO_CONTENT)

        delete.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Delete a resource
            description:
                Delete a resource in this collection.
            responses:
                204:
                    description:
                        Resource deleted successfully.
                404:
                    description:
                        Resource not found.
        """

        return {"_delete": delete}


class ListMixin:
    @classmethod
    def _add_list(
        cls, name: str, verbose_name: str, rest_schemas: data_structures.Schemas, **kwargs
    ) -> dict[str, t.Any]:
        @resource_method("/", methods=["GET"], name="list", pagination="page_number")
        async def list(
            self,
            worker: FlamaWorker,
            order_by: t.Optional[str] = None,
            order_direction: str = "asc",
            **kwargs,
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(rest_schemas.output.schema)]:
            async with worker:
                repository = worker.repositories[self._meta.name]
                return [  # type: ignore[return-value]
                    x async for x in repository.list(order_by=order_by, order_direction=order_direction)
                ]

        list.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                List collection
            description:
                List all resources in this collection.
            responses:
                200:
                    description:
                        Resources list.
        """

        return {"_list": list}


class ReplaceMixin:
    @classmethod
    def _add_replace(
        cls,
        name: str,
        verbose_name: str,
        rest_schemas: data_structures.Schemas,
        rest_model: data_structures.Model,
        **kwargs,
    ) -> dict[str, t.Any]:
        @resource_method("/", methods=["PUT"], name="replace")
        async def replace(
            self,
            worker: FlamaWorker,
            resources: t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(rest_schemas.input.schema)],
        ) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(rest_schemas.output.schema)]:
            async with worker:
                repository = worker.repositories[self._meta.name]
                await repository.drop()
                try:
                    return await repository.create(*resources)
                except ddd_exceptions.IntegrityError as e:
                    raise exceptions.HTTPException(status_code=HTTPStatus.BAD_REQUEST, detail=str(e))

        replace.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Replace collection
            description:
                Replace all resources in this collection.
            responses:
                200:
                    description:
                        Collection replaced successfully.
                400:
                    description:
                        Wrong input data.
        """

        return {"_replace": replace}


class PartialReplaceMixin:
    @classmethod
    def _add_partial_replace(
        cls,
        name: str,
        verbose_name: str,
        rest_schemas: data_structures.Schemas,
        rest_model: data_structures.Model,
        **kwargs,
    ) -> dict[str, t.Any]:
        @resource_method("/", methods=["PATCH"], name="partial-replace")
        async def partial_replace(
            self,
            worker: FlamaWorker,
            resources: t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(rest_schemas.input.schema)],
        ) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(rest_schemas.output.schema)]:
            async with worker:
                repository = worker.repositories[self._meta.name]
                await repository.drop(
                    rest_model.table.c[rest_model.primary_key.name].in_(
                        [x[rest_model.primary_key.name] for x in resources]
                    )
                )
                try:
                    return await repository.create(*resources)
                except ddd_exceptions.IntegrityError as e:
                    raise exceptions.HTTPException(status_code=HTTPStatus.BAD_REQUEST, detail=str(e))

        partial_replace.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Partially replace collection
            description:
                Replace and create resources in this collection.
            responses:
                200:
                    description:
                        Collection replaced successfully.
                400:
                    description:
                        Wrong input data.
        """

        return {"_partial_replace": partial_replace}


class DropMixin:
    @classmethod
    def _add_drop(cls, name: str, verbose_name: str, **kwargs) -> dict[str, t.Any]:
        @resource_method("/", methods=["DELETE"], name="drop")
        async def drop(
            self, worker: FlamaWorker
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(schemas.schemas.DropCollection)]:
            async with worker:
                repository = worker.repositories[self._meta.name]
                result = await repository.drop()

            return http.APIResponse(  # type: ignore[return-value]
                schema=schemas.schemas.DropCollection, content={"deleted": result}, status_code=HTTPStatus.NO_CONTENT
            )

        drop.__doc__ = f"""
            tags:
                - {verbose_name}
            summary:
                Drop collection
            description:
                Drop all resources in this collection.
            responses:
                204:
                    description:
                        Collection dropped successfully.
        """

        return {"_drop": drop}


class CRUDResourceType(
    RESTResourceType,
    CreateMixin,
    RetrieveMixin,
    UpdateMixin,
    PartialUpdateMixin,
    DeleteMixin,
    ListMixin,
    ReplaceMixin,
    PartialReplaceMixin,
    DropMixin,
):
    METHODS = ("create", "retrieve", "update", "partial_update", "delete", "list", "replace", "partial_replace", "drop")

    @staticmethod
    def _is_abstract(namespace: dict[str, t.Any]) -> bool:
        return namespace.get("__module__") == "flama.resources.crud" and namespace.get("__qualname__") == "CRUDResource"


class CRUDResource(RESTResource, metaclass=CRUDResourceType):
    ...



import dataclasses
import typing as t

__all__ = [
    "_T_Field",
    "_T_Schema",
    "SchemaType",
    "SchemaMetadata",
    "get_schema_metadata",
    "is_schema",
    "is_schema_partial",
    "is_schema_multiple",
]

_T_Field = t.TypeVar("_T_Field")
_T_Schema = t.TypeVar("_T_Schema")

SchemaType = dict[str, t.Any]


@dataclasses.dataclass(frozen=True)
class SchemaMetadata:
    schema: t.Any
    partial: bool = False
    multiple: bool = False


def get_schema_metadata(obj: t.Any) -> SchemaMetadata:
    return getattr(obj, "__metadata__", [None])[0]


def is_schema(obj: t.Any) -> bool:
    return isinstance(get_schema_metadata(obj), SchemaMetadata)


def is_schema_partial(obj: t.Any) -> bool:
    return is_schema(obj) and get_schema_metadata(obj).partial


def is_schema_multiple(obj: t.Any) -> bool:
    return is_schema(obj) and get_schema_metadata(obj).multiple



import dataclasses
import typing as t
import urllib.parse

__all__ = ["Secret", "URL"]


class Secret:
    """Secret object.

    It is used to hide sensitive data in logs and exceptions. It is recommended to use this class for all sensitive
    data. For example: passwords, tokens, etc.

    This class is not meant to be used for encryption or security purposes.
    """

    def __init__(self, value: str):
        """Secret object.

        :param value: Sensitive data.
        """
        self._value = value

    def __repr__(self) -> str:
        return "Secret('*****')"

    def __str__(self) -> str:
        return self._value

    def __bool__(self) -> bool:
        return bool(self._value)

    def __eq__(self, other: t.Any) -> bool:
        return self._value == other


@dataclasses.dataclass(frozen=True)
class URL:
    """URL object. It is used to parse and build URLs."""

    scheme: str
    netloc: str
    path: str
    params: str
    query: str
    fragment: str

    def __init__(self, url: str):
        """URL object.

        :param url: URL string to be parsed.
        :param kwargs: Individual components to replace those parsed from string.
        """
        parsed_url = urllib.parse.urlparse(url)
        object.__setattr__(self, "scheme", parsed_url.scheme)
        object.__setattr__(self, "netloc", parsed_url.netloc)
        object.__setattr__(self, "path", parsed_url.path)
        object.__setattr__(self, "params", parsed_url.params)
        object.__setattr__(self, "query", parsed_url.query)
        object.__setattr__(self, "fragment", parsed_url.fragment)

    @property
    def components(self) -> dict[str, t.Optional[str]]:
        """URL components map.

        :return: Components.
        """
        return dataclasses.asdict(self)

    @property
    def url(self) -> str:
        """Build URL string.

        :return: URL string.
        """
        return str(urllib.parse.urlunparse(tuple(self.components.values())))

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return f"URL('{self.url}')"

    def __eq__(self, other: t.Any) -> bool:
        return (isinstance(other, URL) and self.components == other.components) or (
            isinstance(other, str) and self.url == other
        )



import asyncio
import functools
import importlib.metadata
import logging
import typing as t
from types import TracebackType

import httpx

from flama import types
from flama.applications import Flama

__all__ = ["Client", "LifespanContextManager"]

logger = logging.getLogger(__name__)


class LifespanContextManager:
    def __init__(self, app: "Flama", timeout: float = 60.0):
        self.app = app
        self.timeout = timeout
        self._startup_complete = asyncio.Event()
        self._shutdown_complete = asyncio.Event()
        self._receive_queue = asyncio.Queue(maxsize=2)
        self._exception: t.Optional[BaseException] = None

    async def _startup(self) -> None:
        await self._receive_queue.put(types.Message({"type": "lifespan.startup"}))
        await asyncio.wait_for(self._startup_complete.wait(), timeout=self.timeout)
        if self._exception:
            raise self._exception

    async def _shutdown(self) -> None:
        await self._receive_queue.put(types.Message({"type": "lifespan.shutdown"}))
        await asyncio.wait_for(self._shutdown_complete.wait(), timeout=self.timeout)

    async def _receive(self) -> types.Message:
        return await self._receive_queue.get()

    async def _send(self, message: types.Message) -> None:
        if message["type"] == "lifespan.startup.complete":
            self._startup_complete.set()
        elif message["type"] == "lifespan.shutdown.complete":
            self._shutdown_complete.set()

    async def _app_task(self) -> None:
        scope = types.Scope({"type": "lifespan"})

        try:
            await self.app(scope, self._receive, self._send)
        except BaseException as exc:
            self._exception = exc
            self._startup_complete.set()
            self._shutdown_complete.set()

            raise

    async def __aenter__(self) -> "LifespanContextManager":
        task = asyncio.create_task(self._app_task())

        try:
            await self._startup()
        except BaseException:
            raise
        finally:
            await task

        return self

    async def __aexit__(
        self,
        exc_type: t.Optional[type[BaseException]] = None,
        exc_value: t.Optional[BaseException] = None,
        traceback: t.Optional[TracebackType] = None,
    ):
        task = asyncio.create_task(self._app_task())

        try:
            await self._shutdown()
        except BaseException:
            raise
        finally:
            await task


class Client(httpx.AsyncClient):
    """A client for interacting with a Flama application either remote or local.

    This client can handle a local python object:
    >>> client = Client(app=Flama())

    Or connect to a remote API:
    >>> client = Client(base_url="https://foo.bar")

    Or generate a Flama application based on a set of flm model files:
    >>> client = Client(models=[("foo", "/foo/", "model_foo.flm"), ("bar", "/bar/", "model_bar.flm")])

    For initializing the application it's required to use it as an async context manager:
    >>> async with Client(app=Flama()) as client:
    >>>     client.post(...)
    """

    def __init__(
        self,
        /,
        app: t.Optional["Flama"] = None,
        models: t.Optional[t.Sequence[tuple[str, str, str]]] = None,
        **kwargs,
    ):
        self.models: t.Optional[dict[str, str]] = None

        if models:
            app = Flama() if not app else app
            for name, url, path in models:
                app.models.add_model(url, path, name)

            self.models = {m[0]: m[1] for m in models or {}}

        self.lifespan = LifespanContextManager(app) if app else None
        self.app = app

        kwargs.setdefault("transport", httpx.ASGITransport(app=app) if app else None)  # type: ignore
        kwargs.setdefault("base_url", "http://localapp")
        kwargs["headers"] = {"user-agent": f"flama/{importlib.metadata.version('flama')}", **kwargs.get("headers", {})}

        super().__init__(**kwargs)

    async def __aenter__(self) -> "Client":
        if self.lifespan:
            await self.lifespan.__aenter__()
        await super().__aenter__()

        return self

    async def __aexit__(
        self,
        exc_type: t.Optional[type[BaseException]] = None,
        exc_value: t.Optional[BaseException] = None,
        traceback: t.Optional[TracebackType] = None,
    ):
        await super().__aexit__(exc_type, exc_value, traceback)
        if self.lifespan:
            await self.lifespan.__aexit__(exc_type, exc_value, traceback)

    def model_request(self, model: str, method: str, url: str, **kwargs) -> t.Awaitable[httpx.Response]:
        assert self.models, "No models found for request."
        return self.request(method, f"{self.models[model].rstrip('/')}{url}", **kwargs)

    model_inspect = functools.partialmethod(model_request, method="GET", url="/")
    model_predict = functools.partialmethod(model_request, method="POST", url="/predict/")



import abc
import typing as t

from flama import compat
from flama.schemas.types import _T_Field, _T_Schema
from flama.types import JSONSchema

__all__ = ["Adapter"]


class Adapter(t.Generic[_T_Schema, _T_Field], metaclass=abc.ABCMeta):
    DEFAULT_SCHEMA_NAME = "Schema"

    @abc.abstractmethod
    def build_field(
        self,
        name: str,
        type_: type,
        nullable: bool = False,
        required: bool = True,
        default: t.Any = None,
        multiple: bool = False,
        **kwargs: t.Any,
    ) -> _T_Field:
        ...

    @t.overload
    @abc.abstractmethod
    def build_schema(self, *, name: t.Optional[str] = None, fields: dict[str, t.Any]) -> t.Any:
        ...

    @t.overload
    @abc.abstractmethod
    def build_schema(self, *, name: t.Optional[str] = None, schema: t.Any) -> t.Any:
        ...

    @t.overload
    @abc.abstractmethod
    def build_schema(self, *, name: t.Optional[str] = None, schema: t.Any, partial: bool) -> t.Any:
        ...

    @t.overload
    @abc.abstractmethod
    def build_schema(
        self, *, name: t.Optional[str] = None, schema: t.Any, fields: t.Optional[dict[str, t.Any]]
    ) -> t.Any:
        ...

    @abc.abstractmethod
    def build_schema(
        self,
        *,
        name: t.Optional[str] = None,
        schema: t.Optional[t.Any] = None,
        fields: t.Optional[dict[str, t.Any]] = None,
        partial: bool = False,
    ) -> t.Any:
        ...

    @abc.abstractmethod
    def validate(self, schema: t.Any, values: dict[str, t.Any], *, partial: bool = False) -> dict[str, t.Any]:
        ...

    @abc.abstractmethod
    def load(self, schema: t.Any, value: dict[str, t.Any]) -> _T_Schema:
        ...

    @abc.abstractmethod
    def dump(self, schema: t.Any, value: dict[str, t.Any]) -> dict[str, t.Any]:
        ...

    @t.overload
    @abc.abstractmethod
    def name(self, schema: t.Any) -> str:
        ...

    @t.overload
    @abc.abstractmethod
    def name(self, schema: t.Any, *, prefix: str) -> str:
        ...

    @abc.abstractmethod
    def name(self, schema: t.Any, *, prefix: t.Optional[str] = None) -> str:
        ...

    @abc.abstractmethod
    def to_json_schema(self, schema: t.Any) -> JSONSchema:
        ...

    @abc.abstractmethod
    def unique_schema(self, schema: t.Any) -> t.Any:
        ...

    @abc.abstractmethod
    def schema_fields(self, schema: t.Any) -> dict[str, t.Any]:
        ...

    @abc.abstractmethod
    def is_schema(self, obj: t.Any) -> compat.TypeGuard[t.Any]:  # PORT: Replace compat when stop supporting 3.9
        ...

    @abc.abstractmethod
    def is_field(self, obj: t.Any) -> compat.TypeGuard[t.Any]:  # PORT: Replace compat when stop supporting 3.9
        ...



import abc
import dataclasses
import inspect
import typing as t

from flama.injection.exceptions import ComponentNotFound
from flama.injection.types import BUILTIN_TYPES

if t.TYPE_CHECKING:
    from flama.injection.components import Component, Components

__all__ = ["Return", "Parameter", "Resolver"]


EMPTY = t.NewType("EMPTY", object)
ROOT_NAME = "_root"


@dataclasses.dataclass(frozen=True)
class Parameter:
    """A parameter that is part of a function signature."""

    empty = EMPTY

    name: str
    annotation: t.Any = EMPTY
    default: t.Any = EMPTY

    @classmethod
    def from_parameter(cls, parameter: inspect.Parameter) -> "Parameter":
        return cls(
            name=parameter.name,
            annotation=parameter.annotation if parameter.annotation is not parameter.empty else EMPTY,
            default=parameter.default if parameter.default is not parameter.empty else EMPTY,
        )


@dataclasses.dataclass(frozen=True)
class Return(Parameter):
    """A special parameter that represents the return value of a function."""

    name: str = dataclasses.field(init=False, default="_return")
    default: t.Any = dataclasses.field(init=False, default=EMPTY)

    @classmethod
    def from_return_annotation(cls, return_annotation: t.Any) -> "Return":
        return Return(return_annotation if return_annotation is not inspect.Signature.empty else EMPTY)


@dataclasses.dataclass(frozen=True)
class ResolutionNode(abc.ABC):
    """A single node in the dependencies tree."""

    name: str
    parameter: Parameter
    nodes: list["ResolutionNode"]

    @abc.abstractmethod
    async def value(self, context: dict[str, t.Any]) -> t.Any:
        ...

    def components(self) -> list[tuple[str, "Component"]]:
        return []

    def parameters(self) -> list[Parameter]:
        return []

    def context(self) -> list[tuple[str, Parameter]]:
        return []


@dataclasses.dataclass(frozen=True)
class ComponentNode(ResolutionNode):
    """A node that represents a parameter that is resolved by component."""

    component: "Component"

    async def value(self, context: dict[str, t.Any]) -> t.Any:
        kwargs = {node.name: await node.value(context) for node in self.nodes}

        return await self.component(**kwargs)

    def components(self) -> list[tuple[str, "Component"]]:
        return [(self.name, self.component), *[x for node in self.nodes for x in node.components()]]

    def context(self) -> list[tuple[str, Parameter]]:
        return [x for node in self.nodes for x in node.context()]

    def parameters(self) -> list[Parameter]:
        return [x for node in self.nodes for x in node.parameters()]


@dataclasses.dataclass(frozen=True)
class ContextNode(ResolutionNode):
    """A node that represents a parameter that is resolved by context."""

    async def value(self, context: dict[str, t.Any]) -> t.Any:
        return context[self.parameter.name]

    def context(self) -> list[tuple[str, Parameter]]:
        return [(self.name, self.parameter)]


@dataclasses.dataclass(frozen=True)
class ParameterNode(ResolutionNode):
    """A node that represents a parameter that is resolved by another parameter."""

    async def value(self, context: dict[str, t.Any]) -> t.Any:
        return self.parameter

    def parameters(self) -> list[Parameter]:
        return [self.parameter]


@dataclasses.dataclass(frozen=True)
class ResolutionTree:
    """Dependencies tree."""

    root: ResolutionNode
    context: list[tuple[str, Parameter]] = dataclasses.field(init=False)
    parameters: list[Parameter] = dataclasses.field(init=False)
    components: list[tuple[str, "Component"]] = dataclasses.field(init=False)

    def __post_init__(self):
        object.__setattr__(self, "context", self.root.context())
        object.__setattr__(self, "parameters", self.root.parameters())
        object.__setattr__(self, "components", self.root.components())

    @classmethod
    def build(cls, parameter: Parameter, context_types: dict[t.Any, str], components: "Components") -> "ResolutionTree":
        return cls(root=cls._build_node(parameter, context_types, components))

    @classmethod
    def _build_node(
        cls,
        parameter: Parameter,
        context_types: dict[t.Any, str],
        components: "Components",
        parent: t.Optional[Parameter] = None,
    ) -> ResolutionNode:
        assert parameter.name not in ("self", "cls")

        # Check if the parameter annotation exists in context types.
        if parameter.annotation in context_types:
            return ContextNode(
                parameter.name, Parameter(context_types[parameter.annotation], parameter.annotation), nodes=[]
            )

        # The 'Parameter' annotation can be used to get the parameter itself, so it is stored as a constant.
        if parameter.annotation is Parameter:
            assert parent is not None, "A root function cannot define an argument with Parameter type"

            return ParameterNode(parameter.name, parent, nodes=[])

        # Look for a component that can handles the parameter.
        try:
            component = components.find_handler(parameter)
        except ComponentNotFound:
            # There is no component that can handles this parameter, and it is a builtin type, so it'll be expected as
            # part of the building context values.
            if parameter.annotation in BUILTIN_TYPES:
                return ContextNode(parameter.name, parameter, nodes=[])

            raise
        else:
            # There is a component that can handles the parameter
            try:
                nodes = [
                    cls._build_node(p, context_types, components, parent=parameter)
                    for p in component.signature().values()
                    if p.name not in ("self", "cls")
                ]
                return ComponentNode(parameter.name, parameter, nodes, component)
            except ComponentNotFound as e:
                if e.component is None:
                    raise ComponentNotFound(e.parameter, component=component) from None

                raise  # pragma: no cover

    async def value(self, context: dict[str, t.Any]) -> t.Any:
        return await self.root.value(context)


class ResolutionCache(t.Mapping[int, ResolutionTree]):
    """A cache for resolution trees."""

    def __init__(self):
        self._data = {}

    def _can_cache(self, value: ResolutionTree) -> bool:
        return isinstance(value.root, ComponentNode) and not value.root.component.use_parameter

    def __setitem__(self, key: int, value: ResolutionTree) -> None:
        if not self._can_cache(value):
            raise ValueError(f"Resolution {value} cannot be cached")

        self._data.__setitem__(key, value)

    def __getitem__(self, key: int) -> t.Any:
        return self._data.__getitem__(key)

    def __eq__(self, other: object) -> bool:
        return self._data.__eq__(other)

    def __iter__(self) -> t.Iterator:
        return self._data.__iter__()

    def __len__(self) -> int:
        return self._data.__len__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._data.__repr__()})"


class Resolver:
    """Provides a way to inspect a parameter and build it dependencies tree."""

    def __init__(self, context_types: dict[str, type], components: "Components"):
        """Initialize a new resolver.

        The context types are used to determine if a parameter is a context value or not. The components registry will
        be used to determine if a parameter can be handled by a component or not.

        :param context_types: A dictionary that maps a context type name to a type.
        :param components: A components registry.
        """
        self.context_types = {v: k for k, v in context_types.items()}
        self.components = components
        self._cache = ResolutionCache()

    def resolve(self, parameter: Parameter) -> ResolutionTree:
        """Build a resolution tree for the given parameter.

        :param parameter: The parameter to be resolved.
        :return: A resolution tree.
        """
        key = hash(parameter.annotation)
        try:
            resolution = self._cache[key]
        except KeyError:
            resolution = ResolutionTree.build(parameter, self.context_types, self.components)
            try:
                self._cache[key] = resolution
            except ValueError:
                ...

        return resolution



import abc
import dataclasses
import inspect
import logging
import typing as t
from pathlib import Path

import starlette.exceptions

from flama import concurrency, exceptions, http, types, websockets
from flama.debug.data_structures import ErrorContext, NotFoundContext

if t.TYPE_CHECKING:
    from flama.debug.types import Handler

logger = logging.getLogger(__name__)

__all__ = ["ServerErrorMiddleware", "ExceptionMiddleware"]

TEMPLATES_PATH = Path(__file__).parents[1].resolve() / "templates" / "debug"


class BaseErrorMiddleware:
    def __init__(self, app: types.App, debug: bool = False) -> None:
        self.app = app
        self.debug = debug

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        response_started = False

        async def sender(message: types.Message) -> None:
            nonlocal response_started, send

            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await concurrency.run(self.app, scope, receive, sender)
        except Exception as exc:
            if scope["type"] in ("http", "websocket"):
                await self.process_exception(scope, receive, send, exc, response_started)
            else:
                raise

    @abc.abstractmethod
    async def process_exception(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: Exception, response_started: bool
    ) -> None:
        ...


class ServerErrorMiddleware(BaseErrorMiddleware):
    def _get_handler(self, scope: types.Scope) -> "Handler":
        if scope["type"] == "http":
            return self.debug_handler if self.debug else self.error_handler

        return self.noop_handler

    async def process_exception(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: Exception, response_started: bool
    ) -> None:
        logger.exception("Unhandled exception '%s' for scope %s", exc, scope)

        handler = self._get_handler(scope)

        logger.debug("Exception handler: %s", handler)

        response = handler(scope, receive, send, exc)

        if response and concurrency.is_async(response) and not response_started:
            await response(scope, receive=receive, send=send)

        # We always continue to raise the exception.
        # This allows servers to log the error, or test clients to optionally raise the error within the test case.
        raise exc

    def debug_handler(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: Exception
    ) -> http.Response:
        request = http.Request(scope)
        accept = request.headers.get("accept", "")

        if "text/html" in accept:
            return http._FlamaTemplateResponse(
                "debug/error_500.html", context=dataclasses.asdict(ErrorContext.build(request, exc)), status_code=500
            )
        return http.PlainTextResponse("Internal Server Error", status_code=500)

    def error_handler(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: Exception
    ) -> http.Response:
        return http.PlainTextResponse("Internal Server Error", status_code=500)

    def noop_handler(self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: Exception) -> None:
        ...


class ExceptionMiddleware(BaseErrorMiddleware):
    def __init__(self, app: types.App, handlers: t.Optional[t.Mapping[t.Any, "Handler"]] = None, debug: bool = False):
        super().__init__(app, debug)
        handlers = handlers or {}
        self._status_handlers: dict[int, Handler] = {
            status_code: handler for status_code, handler in handlers.items() if isinstance(status_code, int)
        }
        self._exception_handlers: dict[type[Exception], Handler] = {
            **{e: handler for e, handler in handlers.items() if inspect.isclass(e) and issubclass(e, Exception)},
            exceptions.NotFoundException: self.not_found_handler,
            exceptions.MethodNotAllowedException: self.method_not_allowed_handler,
            starlette.exceptions.HTTPException: self.http_exception_handler,
            starlette.exceptions.WebSocketException: self.websocket_exception_handler,
        }

    def add_exception_handler(
        self,
        handler: "Handler",
        status_code: t.Optional[int] = None,
        exc_class: t.Optional[type[Exception]] = None,
    ) -> None:
        if status_code is None and exc_class is None:
            raise ValueError("Status code or exception class must be defined")

        if status_code is not None:
            self._status_handlers[status_code] = handler

        if exc_class is not None:
            self._exception_handlers[exc_class] = handler

    def _get_handler(self, exc: Exception) -> "Handler":
        if isinstance(exc, starlette.exceptions.HTTPException) and exc.status_code in self._status_handlers:
            return self._status_handlers[exc.status_code]
        else:
            try:
                return next(
                    self._exception_handlers[cls] for cls in type(exc).__mro__ if cls in self._exception_handlers
                )
            except StopIteration:
                raise exc

    async def process_exception(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: Exception, response_started: bool
    ) -> None:
        logger.debug("Handled exception '%s' for scope %s", exc, scope, exc_info=True)

        handler = self._get_handler(exc)

        logger.debug("Exception handler: %s", handler)

        if response_started:
            raise RuntimeError("Caught handled exception, but response already started.") from exc

        response = await concurrency.run(handler, scope, receive, send, exc)

        if response and concurrency.is_async(response):
            await response(scope, receive, send)

    def http_exception_handler(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: exceptions.HTTPException
    ) -> http.Response:
        if exc.status_code in {204, 304}:
            return http.Response(status_code=exc.status_code, headers=exc.headers)

        request = http.Request(scope, receive=receive)
        accept = request.headers.get("accept", "")

        if self.debug and exc.status_code == 404 and "text/html" in accept:
            return http._FlamaTemplateResponse(
                template="debug/error_404.html",
                context=dataclasses.asdict(NotFoundContext.build(request, scope["app"])),
                status_code=404,
            )

        return http.APIErrorResponse(detail=exc.detail, status_code=exc.status_code, exception=exc)

    async def websocket_exception_handler(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: exceptions.WebSocketException
    ) -> None:
        websocket = websockets.WebSocket(scope, receive=receive, send=send)
        await websocket.close(code=exc.code, reason=exc.reason)

    async def not_found_handler(
        self, scope: types.Scope, receive: types.Receive, send: types.Send, exc: exceptions.NotFoundException
    ) -> t.Optional[http.Response]:
        if scope.get("type", "") == "websocket":
            await self.websocket_exception_handler(scope, receive, send, exc=exceptions.WebSocketException(1000))
            return None

        if "app" in scope:
            return self.http_exception_handler(scope, receive, send, exc=exceptions.HTTPException(status_code=404))

        return http.PlainTextResponse("Not Found", status_code=404)

    async def method_not_allowed_handler(
        self,
        scope: types.Scope,
        receive: types.Receive,
        send: types.Send,
        exc: exceptions.MethodNotAllowedException,
    ) -> t.Optional[http.Response]:
        if scope.get("type", "") == "websocket":
            await self.websocket_exception_handler(scope, receive, send, exc=exceptions.WebSocketException(1000))
            return None

        return self.http_exception_handler(
            scope,
            receive,
            send,
            exc=exceptions.HTTPException(status_code=405, headers={"Allow": ", ".join(exc.allowed)}),
        )



import dataclasses
import typing as t


@dataclasses.dataclass
class Events:
    """Application events register."""

    startup: list[t.Callable[..., t.Coroutine[t.Any, t.Any, None]]] = dataclasses.field(default_factory=list)
    shutdown: list[t.Callable[..., t.Coroutine[t.Any, t.Any, None]]] = dataclasses.field(default_factory=list)

    def register(self, event: str, handler: t.Callable) -> None:
        """Register a new event.

        :param event: Event type.
        :param handler: Event handler.
        """
        assert event in {f.name for f in dataclasses.fields(self)}, f"Wrong event: {event}."
        getattr(self, event).append(handler)

    @classmethod
    def build(cls, **events: list[t.Callable[..., t.Coroutine[t.Any, t.Any, None]]]) -> "Events":
        """Build events register from dict.

        :param events: Events to register.
        :return: Events instance.
        """
        keys = set(events.keys()) - {f.name for f in dataclasses.fields(cls)}
        assert not keys, f"Wrong event{'s' if len(keys) > 1 else ''}: {', '.join(keys)}."
        return cls(**events)



import typing

__all__ = ["SchemaError", "SchemaParseError", "SchemaValidationError", "SchemaGenerationError"]


class SchemaError(Exception):
    def __init__(self, errors: typing.Union[str, dict[str, typing.Any]], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.errors = errors


class SchemaParseError(SchemaError):
    pass


class SchemaValidationError(SchemaError):
    pass


class SchemaGenerationError(Exception):
    pass



import abc
import logging
import typing as t

from flama import exceptions
from flama.modules import Module

try:
    import sqlalchemy
    from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, AsyncTransaction, create_async_engine

    metadata = sqlalchemy.MetaData()
except Exception:  # pragma: no cover
    raise exceptions.DependencyNotInstalled(
        dependency=exceptions.DependencyNotInstalled.Dependency.sqlalchemy, dependant=__name__
    )


__all__ = ["metadata", "SQLAlchemyModule"]

logger = logging.getLogger(__name__)


class ConnectionManager(abc.ABC):
    """Abstract class for connection managers.

    It will be used to manage the connections and transactions.
    """

    def __init__(self, engine: AsyncEngine) -> None:
        """Initialize the connection manager.

        :param engine: SQLAlchemy engine.
        """
        self._engine = engine

    @abc.abstractmethod
    async def open(self) -> AsyncConnection:
        """Open a new connection to the database.

        :return: Database connection.
        """
        ...

    @abc.abstractmethod
    async def close(self, connection: AsyncConnection) -> None:
        """Close the connection to the database.

        :param connection: Database connection.
        """
        ...

    @abc.abstractmethod
    async def begin(self, connection: AsyncConnection) -> AsyncTransaction:
        """Begin a new transaction.

        :param connection: Database connection to use for the transaction.
        :return: Database transaction.
        """
        ...

    @abc.abstractmethod
    async def end(self, transaction: AsyncTransaction, *, rollback: bool = False) -> None:
        """End a transaction.

        :param transaction: Database transaction.
        :param rollback: If the transaction should be rolled back.
        """
        ...


class SingleConnectionManager(ConnectionManager):
    """Connection manager that uses a single connection and transaction.

    A single connection is opened when requested, and subsequent requests will share this same connection. Once all
    clients finishes with the connection, it will be closed.

    The transaction is similar to the transaction, but it will be created at first request and subsequent requests will
    generate a nested transaction.
    """

    def __init__(self, engine: AsyncEngine) -> None:
        """Initialize the connection manager.

        :param engine: SQLAlchemy engine.
        """
        super().__init__(engine)
        self._connection: t.Optional[AsyncConnection] = None
        self._transaction: t.Optional[AsyncTransaction] = None
        self._connection_clients = 0
        self._transaction_clients = 0

    @property
    def connection(self) -> AsyncConnection:
        """Connection to the database.

        :return: Connection to the database.
        :raises SQLAlchemyError: If the connection is not initialized.
        """
        if not self._connection:
            raise exceptions.SQLAlchemyError("Connection not initialized")

        return self._connection

    @property
    def transaction(self) -> AsyncTransaction:
        """Transaction to the database.

        :return: Transaction to the database.
        :raises SQLAlchemyError: If the transaction is not initialized.
        """
        if not self._transaction:
            raise exceptions.SQLAlchemyError("Transaction not started")

        return self._transaction

    async def open(self) -> AsyncConnection:
        """Open a new connection to the database.

        The first client will open a new connection, and subsequent clients will share this same connection.

        :return: Database connection.
        """
        try:
            connection = self.connection
        except exceptions.SQLAlchemyError:
            self._connection = connection = self._engine.connect()
            await self._connection.__aenter__()

        self._connection_clients += 1
        return connection

    async def close(self, connection: AsyncConnection) -> None:
        """Close the connection to the database.

        If this is the last client, the connection will be closed.

        :param connection: Database connection.
        :raises SQLAlchemyError: If the connection is a different connection from the one opened.
        """
        if connection != self.connection:
            raise exceptions.SQLAlchemyError("Wrong connection")

        self._connection_clients -= 1

        if self._connection_clients == 0:
            await connection.__aexit__(None, None, None)
            self._connection = None

    async def begin(self, connection: AsyncConnection) -> AsyncTransaction:
        """Begin a new transaction.

        The first client will create a new transaction, and subsequent clients will share this same transaction.

        :return: Database transaction.
        :raises SQLAlchemyError: If the connection is a different connection from the one opened.
        """
        if connection != self.connection:
            raise exceptions.SQLAlchemyError("Wrong connection")

        try:
            transaction = self.transaction
        except exceptions.SQLAlchemyError:
            self._transaction = transaction = connection.begin()
            await transaction

        self._transaction_clients += 1
        return transaction

    async def end(self, transaction: AsyncTransaction, *, rollback: bool = False) -> None:
        """End a transaction.

        If this is the last client, the connection will be committed or rolled back.

        :param transaction: Database transaction.
        :param rollback: If the transaction should be rolled back.
        """
        if transaction != self.transaction:
            raise exceptions.SQLAlchemyError("Wrong transaction")

        self._transaction_clients -= 1

        if self._transaction_clients == 0:
            if rollback:
                await transaction.rollback()
            else:
                await transaction.commit()

            self._transaction = None


class MultipleConnectionManager(ConnectionManager):
    """Connection manager that handlers several connections and transactions."""

    def __init__(self, engine: AsyncEngine) -> None:
        """Initialize the connection manager.

        This manager keeps track of the connections and transactions, and it will close the connections when requested.
        If a connection is closed, all their transactions will be commited and finished too.

        :param engine: SQLAlchemy engine.
        """
        super().__init__(engine)
        self._connections: set[AsyncConnection] = set()
        self._transactions: dict[AsyncConnection, AsyncTransaction] = {}

    async def open(self) -> AsyncConnection:
        """Open a new connection to the database.

        :return: Database connection.
        """
        connection = self._engine.connect()
        await connection.start()
        self._connections.add(connection)
        return connection

    async def close(self, connection: AsyncConnection) -> None:
        """Close the connection to the database.

        :param connection: Database connection.
        :raises SQLAlchemyError: If the connection is not initialized.
        """
        if connection not in self._connections:
            raise exceptions.SQLAlchemyError("Connection not initialized")

        if connection in self._transactions:
            await self.end(self._transactions[connection])

        self._connections.remove(connection)
        await connection.close()

    async def begin(self, connection: AsyncConnection) -> AsyncTransaction:
        """Begin a new transaction.

        :param connection: Database connection to use for the transaction.
        :return: Database transaction.
        :raises SQLAlchemyError: If the connection is not initialized.
        """
        if connection not in self._connections:
            raise exceptions.SQLAlchemyError("Connection not initialized")

        if connection in self._transactions:
            raise exceptions.SQLAlchemyError("Transaction already started in this connection")

        transaction = await connection.begin()
        self._transactions[connection] = transaction
        return transaction

    async def end(self, transaction: AsyncTransaction, *, rollback: bool = False) -> None:
        """End a transaction.

        :param transaction: Database transaction.
        :param rollback: If the transaction should be rolled back.
        :raises SQLAlchemyError: If the transaction is not started.
        """
        if transaction.connection not in self._transactions:
            raise exceptions.SQLAlchemyError("Transaction not started")

        del self._transactions[transaction.connection]

        if rollback:
            await transaction.rollback()
        else:
            await transaction.commit()


class SQLAlchemyModule(Module):
    """SQLAlchemy module.

    It will initialize the SQLAlchemy engine and connection manager. It will also provide an interface to handle
    connections and transactions.

    Referring to how connections and transactions are managed, it can work in two modes: single connection and multiple
    connections.
    * Multiple connections (default): It will open a new connection and transaction for each request. It will keep
    track of the connections, and it will close it when requested. Multiple transactions using the same connection is
    also supported.
    * Single connection: It will open a single connection and transaction, and all requests will share this connection
    and transaction. The connection will be closed when all clients finishes. It will create a single transaction and
    new transactions requested will be nested from this one.
    """

    name = "sqlalchemy"

    def __init__(
        self, database: str, single_connection: bool = False, engine_args: t.Optional[dict[str, t.Any]] = None
    ):
        """Initialize the SQLAlchemy module.

        Referring to how connections and transactions are managed, it can work in two modes: single connection and
        multiple connections.
        * Multiple connections (default): It will open a new connection and transaction for each request. It will keep
        track of the connections, and it will close it when requested. Multiple transactions using the same connection
        is also supported.
        * Single connection: It will open a single connection and transaction, and all requests will share this
        connection and transaction. The connection will be closed when all clients finishes. It will create a single
        transaction and new transactions requested will be nested from this one.

        :param database: Database connection string.
        :param single_connection: If the module should work in single connection mode.
        :param engine_args: Arguments to pass to the SQLAlchemy engine.
        :raises ApplicationError: If SQLAlchemy is not installed.
        """
        if not database:
            raise exceptions.ApplicationError("Database connection string must be provided")

        super().__init__()

        self.database = database
        self.metadata: sqlalchemy.MetaData = metadata
        self._engine: t.Optional[AsyncEngine] = None
        self._engine_args = engine_args or {}
        self._connection_manager: t.Optional[ConnectionManager] = None
        self._manager_cls: type[ConnectionManager] = (
            SingleConnectionManager if single_connection else MultipleConnectionManager
        )

    @property
    def engine(self) -> AsyncEngine:
        """SQLAlchemy engine.

        :return: SQLAlchemy engine.
        :raises ApplicationError: If SQLAlchemyModule is not initialized.
        """
        if self._engine is None:
            raise exceptions.ApplicationError("SQLAlchemyModule not initialized")
        return self._engine

    @property
    def connection_manager(self) -> "ConnectionManager":
        """Connection manager.

        :return: Connection manager.
        :raises ApplicationError: If SQLAlchemyModule is not initialized.
        """
        if self._connection_manager is None:
            raise exceptions.ApplicationError("SQLAlchemyModule not initialized")
        return self._connection_manager

    async def open_connection(self) -> AsyncConnection:
        """Open a new connection to the database.

        :return: Database connection.
        """
        return await self.connection_manager.open()

    async def close_connection(self, connection: AsyncConnection) -> None:
        """Close the connection to the database.

        :param connection: Database connection.
        """
        return await self.connection_manager.close(connection)

    async def begin_transaction(self, connection: AsyncConnection) -> AsyncTransaction:
        """Begin a new transaction.

        :param connection: Database connection to use for the transaction.
        :return: Database transaction.
        """
        return await self.connection_manager.begin(connection)

    async def end_transaction(self, transaction: AsyncTransaction, *, rollback: bool = False) -> None:
        """End a transaction.

        :param transaction: Database transaction.
        :param rollback: If the transaction should be rolled back.
        :return: Database transaction.
        """
        return await self.connection_manager.end(transaction, rollback=rollback)

    async def on_startup(self):
        """Initialize the SQLAlchemy engine and connection manager."""
        self._engine = create_async_engine(self.database, **self._engine_args)
        self._connection_manager = self._manager_cls(self._engine)

    async def on_shutdown(self):
        """Close the SQLAlchemy engine and connection manager."""
        await self.engine.dispose()
        self._engine = None
        self._connection_manager = None



import typing as t

from flama import codecs, exceptions, http, schemas, types
from flama.injection import Component, Components
from flama.injection.resolver import Parameter
from flama.negotiation import ContentTypeNegotiator, WebSocketEncodingNegotiator
from flama.routing import BaseRoute
from flama.schemas import Field, Schema, SchemaValidationError

ValidatedPathParams = t.NewType("ValidatedPathParams", dict)
ValidatedQueryParams = t.NewType("ValidatedQueryParams", dict)
ValidatedRequestData = t.NewType("ValidatedRequestData", dict)


class RequestDataComponent(Component):
    def __init__(self):
        self.negotiator = ContentTypeNegotiator(
            [codecs.JSONDataCodec(), codecs.URLEncodedCodec(), codecs.MultiPartCodec()]
        )

    async def resolve(self, request: http.Request) -> types.RequestData:
        content_type = request.headers.get("Content-Type")

        try:
            codec = self.negotiator.negotiate(content_type)
        except exceptions.NoCodecAvailable:
            raise exceptions.HTTPException(415)

        try:
            data = await codec.decode(request)
            return types.RequestData(data) if data else None  # type: ignore
        except exceptions.DecodeError as exc:
            raise exceptions.HTTPException(400, detail=str(exc))


class WebSocketMessageDataComponent(Component):
    def __init__(self):
        self.negotiator = WebSocketEncodingNegotiator([codecs.BytesCodec(), codecs.TextCodec(), codecs.JSONCodec()])

    async def resolve(self, message: types.Message, websocket_encoding: types.Encoding) -> types.Data:
        try:
            codec = self.negotiator.negotiate(websocket_encoding)
            return types.Data(await codec.decode(message))
        except (exceptions.NoCodecAvailable, exceptions.DecodeError):
            raise exceptions.WebSocketException(code=1003)


class ValidatePathParamsComponent(Component):
    async def resolve(
        self, request: http.Request, route: BaseRoute, path_params: types.PathParams
    ) -> ValidatedPathParams:
        fields = [f.field for f in route.parameters.path[request.method].values()]

        try:
            validated = Schema.build(name="ValidationSchema", fields=fields).validate(path_params)
            return ValidatedPathParams({k: v for k, v in path_params.items() if k in validated})
        except SchemaValidationError as exc:
            raise exceptions.ValidationError(detail=exc.errors)


class ValidateQueryParamsComponent(Component):
    def resolve(self, request: http.Request, route: BaseRoute, query_params: types.QueryParams) -> ValidatedQueryParams:
        fields = [f.field for f in route.parameters.query[request.method].values()]

        try:
            validated = Schema.build(name="ValidationSchema", fields=fields).validate(dict(query_params))
            return ValidatedQueryParams({k: v for k, v in query_params.items() if k in validated})
        except SchemaValidationError as exc:
            raise exceptions.ValidationError(detail=exc.errors)


class ValidateRequestDataComponent(Component):
    def resolve(self, request: http.Request, route: BaseRoute, data: types.RequestData) -> ValidatedRequestData:
        body_param = route.parameters.body[request.method]

        assert (
            body_param is not None
        ), f"Body schema parameter not defined for route '{route}' and method '{request.method}'"

        try:
            return ValidatedRequestData(body_param.schema.validate(dict(data)))
        except SchemaValidationError as exc:  # noqa: safety net, just should not happen
            raise exceptions.ValidationError(detail=exc.errors)


class PrimitiveParamComponent(Component):
    def can_handle_parameter(self, parameter: Parameter):
        return Field.is_http_valid_type(parameter.annotation)

    def resolve(self, parameter: Parameter, path_params: ValidatedPathParams, query_params: ValidatedQueryParams):
        params = path_params if (parameter.name in path_params) else query_params

        try:
            params = Schema.build(name="ValidationSchema", fields=[Field.from_parameter(parameter)]).validate(params)
        except SchemaValidationError as exc:  # noqa: safety net, just should not happen
            raise exceptions.ValidationError(detail=exc.errors)
        return params.get(parameter.name, parameter.default)


class CompositeParamComponent(Component):
    def can_handle_parameter(self, parameter: Parameter):
        schema = (
            t.get_args(parameter.annotation)[0] if t.get_origin(parameter.annotation) == list else parameter.annotation
        )
        return schemas.is_schema(schema)

    def resolve(self, parameter: Parameter, request: http.Request, route: BaseRoute, data: types.RequestData):
        body_param = route.parameters.body[request.method]

        assert (
            body_param is not None
        ), f"Body schema parameter not defined for route '{route}' and method '{request.method}'"

        try:
            return body_param.schema.validate(data, partial=schemas.is_schema_partial(parameter.annotation))
        except SchemaValidationError as exc:  # noqa: safety net, just should not happen
            raise exceptions.ValidationError(detail=exc.errors)


VALIDATION_COMPONENTS = Components(
    [
        RequestDataComponent(),
        WebSocketMessageDataComponent(),
        ValidatePathParamsComponent(),
        ValidateQueryParamsComponent(),
        ValidateRequestDataComponent(),
        PrimitiveParamComponent(),
        CompositeParamComponent(),
    ]
)



import typing as t

from flama.codecs.base import HTTPCodec

if t.TYPE_CHECKING:
    from flama import http

__all__ = ["URLEncodedCodec"]


class URLEncodedCodec(HTTPCodec):
    media_type = "application/x-www-form-urlencoded"

    async def decode(self, item: "http.Request", **options):
        return await item.form() or None



import typing as t

from flama import exceptions
from flama.ddd import exceptions as ddd_exceptions
from flama.ddd.repositories import BaseRepository

try:
    import sqlalchemy
    import sqlalchemy.exc as sqlalchemy_exceptions
    from sqlalchemy.ext.asyncio import AsyncConnection
except Exception:  # pragma: no cover
    raise exceptions.DependencyNotInstalled(
        dependency=exceptions.DependencyNotInstalled.Dependency.sqlalchemy, dependant=__name__
    )


__all__ = ["SQLAlchemyRepository", "SQLAlchemyTableManager", "SQLAlchemyTableRepository"]


class SQLAlchemyRepository(BaseRepository):
    """Base class for SQLAlchemy repositories. It provides a connection to the database."""

    def __init__(self, connection: AsyncConnection, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection = connection

    def __eq__(self, other):
        return isinstance(other, SQLAlchemyRepository) and self._connection == other._connection


class SQLAlchemyTableManager:
    def __init__(self, table: sqlalchemy.Table, connection: AsyncConnection):  # type: ignore
        self._connection = connection
        self.table = table
        self.resource = table.name

    def __eq__(self, other):
        return (
            isinstance(other, SQLAlchemyTableManager)
            and self._connection == other._connection
            and self.table == other.table
        )

    async def create(self, *data: dict[str, t.Any]) -> list[dict[str, t.Any]]:
        """Creates new elements in the table.

        If the element already exists, it raises an `IntegrityError`. If the element is created, it returns
        the primary key of the element.

        :param data: The data to create the elements.
        :return: The created elements.
        :raises IntegrityError: If the element already exists or cannot be inserted.
        """
        try:
            result = await self._connection.execute(sqlalchemy.insert(self.table).values(data).returning(self.table))
        except sqlalchemy_exceptions.IntegrityError:
            raise ddd_exceptions.IntegrityError(resource=self.resource)
        return [dict[str, t.Any](element._asdict()) for element in result]

    async def retrieve(self, *clauses, **filters) -> dict[str, t.Any]:
        """Retrieves an element from the table.

        If the element does not exist, it raises a `NotFoundError`. If more than one element is found, it raises a
        `MultipleRecordsError`. If the element is found, it returns the element.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements
        using exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Filter example: `id=1`

        :return: The element.
        :raises NotFoundError: If the element does not exist.
        :raises MultipleRecordsError: If more than one element is found.
        """
        query = self._filter_query(sqlalchemy.select(self.table), *clauses, **filters)

        try:
            element = (await self._connection.execute(query)).one()
        except sqlalchemy_exceptions.NoResultFound:
            raise ddd_exceptions.NotFoundError(resource=self.resource)
        except sqlalchemy_exceptions.MultipleResultsFound:
            raise ddd_exceptions.MultipleRecordsError(resource=self.resource)

        return dict[str, t.Any](element._asdict())

    async def update(self, data: dict[str, t.Any], *clauses, **filters) -> list[dict[str, t.Any]]:
        """Updates elements in the table.

        Using clauses and filters, it filters the elements to update. If no clauses or filters are given, it updates
        all the elements in the table.

        :param data: The data to update the elements.
        :return: The updated elements.
        :raises IntegrityError: If the elements cannot be updated.
        """
        query = (
            self._filter_query(sqlalchemy.update(self.table), *clauses, **filters).values(**data).returning(self.table)
        )

        try:
            result = await self._connection.execute(query)
        except sqlalchemy_exceptions.IntegrityError:
            raise ddd_exceptions.IntegrityError(resource=self.resource)

        return [dict[str, t.Any](element._asdict()) for element in result]

    async def delete(self, *clauses, **filters) -> None:
        """Delete an element from the table.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements using
        exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Filter example: `id=1`

        :param clauses: Clauses to filter the elements.
        :param filters: Filters to filter the elements.
        :raises NotFoundError: If the element does not exist.
        :raises MultipleRecordsError: If more than one element is found.
        """
        await self.retrieve(*clauses, **filters)

        query = self._filter_query(sqlalchemy.delete(self.table), *clauses, **filters)

        await self._connection.execute(query)

    async def list(
        self, *clauses, order_by: t.Optional[str] = None, order_direction: str = "asc", **filters
    ) -> t.AsyncIterable[dict[str, t.Any]]:
        """Lists all the elements in the table.

        If no elements are found, it returns an empty list. If no clauses or filters are given, it returns all the
        elements in the repository.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements using
        exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Order example: `order_by="id", order_direction="desc"`
        Filter example: `id=1`

        :param clauses: Clauses to filter the elements.
        :param order_by: Column to order the elements.
        :param order_direction: Direction to order the elements, either `asc` or `desc`.
        :param filters: Filters to filter the elements.
        :return: Async iterable of the elements.
        """
        query = self._filter_query(sqlalchemy.select(self.table), *clauses, **filters)

        if order_by:
            query = query.order_by(
                self.table.c[order_by].desc() if order_direction == "desc" else self.table.c[order_by]
            )

        result = await self._connection.stream(query)

        async for row in result:
            yield dict[str, t.Any](row._asdict())

    async def drop(self, *clauses, **filters) -> int:
        """Drops elements in the table.

        Returns the number of elements dropped. If no clauses or filters are given, it deletes all the elements in the
        table.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements using
        exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Filter example: `id=1`

        :param clauses: Clauses to filter the elements.
        :param filters: Filters to filter the elements.
        :return: The number of elements dropped.
        """
        query = self._filter_query(sqlalchemy.delete(self.table), *clauses, **filters)

        result = await self._connection.execute(query)

        return result.rowcount

    def _filter_query(self, query, *clauses, **filters):
        """Filters a query using clauses and filters.

        Returns the filtered query. If no clauses or filters are given, it returns the query without any applying
        filter.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements using
        exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Filter example: `id=1`

        :param query: The query to filter.
        :param clauses: Clauses to filter the elements.
        :param filters: Filters to filter the elements.
        :return: The filtered query.
        """
        where_clauses = tuple(clauses) + tuple(self.table.c[k] == v for k, v in filters.items())

        if where_clauses:
            query = query.where(sqlalchemy.and_(*where_clauses))

        return query


class SQLAlchemyTableRepository(SQLAlchemyRepository):
    _table: t.ClassVar[sqlalchemy.Table]  # type: ignore

    def __init__(self, connection: AsyncConnection, *args, **kwargs):
        super().__init__(connection, *args, **kwargs)
        self._table_manager = SQLAlchemyTableManager(self._table, connection)

    def __eq__(self, other):
        return isinstance(other, SQLAlchemyTableRepository) and self._table == other._table and super().__eq__(other)

    async def create(self, *data: dict[str, t.Any]) -> list[dict[str, t.Any]]:
        """Creates new elements in the repository.

        If the element already exists, it raises an `exceptions.IntegrityError`. If the element is created, it returns
        the primary key of the element.

        :param data: The data to create the element.
        :return: The primary key of the created element.
        :raises IntegrityError: If the element already exists or cannot be inserted.
        """
        return await self._table_manager.create(*data)

    async def retrieve(self, *clauses, **filters) -> dict[str, t.Any]:
        """Retrieves an element from the repository.

        If the element does not exist, it raises a `NotFoundError`. If more than one element is found, it raises a
        `MultipleRecordsError`. If the element is found, it returns the element.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements
        using exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Filter example: `id=1`

        :param clauses: Clauses to filter the elements.
        :param filters: Filters to filter the elements.
        :return: The element.
        :raises NotFoundError: If the element does not exist.
        :raises MultipleRecordsError: If more than one element is found.
        """
        return await self._table_manager.retrieve(*clauses, **filters)

    async def update(self, data: dict[str, t.Any], *clauses, **filters) -> list[dict[str, t.Any]]:
        """Updates an element in the repository.

        If the element does not exist, it raises a `NotFoundError`. If the element is updated, it returns the updated
        element.

        :param data: The data to update the element.
        :param clauses: Clauses to filter the elements.
        :param filters: Filters to filter the elements.
        :return: The updated elements.
        :raises IntegrityError: If the elements cannot be updated.
        """
        return await self._table_manager.update(data, *clauses, **filters)

    async def delete(self, *clauses, **filters) -> None:
        """Deletes an element from the repository.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements
        using exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Filter example: `id=1`

        :param id: The primary key of the element.
        :param clauses: Clauses to filter the elements.
        :param filters: Filters to filter the elements.
        :raises NotFoundError: If the element does not exist.
        :raises MultipleRecordsError: If more than one element is found.
        """
        return await self._table_manager.delete(*clauses, **filters)

    def list(
        self, *clauses, order_by: t.Optional[str] = None, order_direction: str = "asc", **filters
    ) -> t.AsyncIterable[dict[str, t.Any]]:
        """Lists all the elements in the repository.

        Lists all the elements in the repository that match the clauses and filters. If no clauses or filters are given,
        it returns all the elements in the repository. If no elements are found, it returns an empty list.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements using
        exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Order example: `order_by="id", order_direction="desc"`
        Filter example: `id=1`

        :param clauses: Clauses to filter the elements.
        :param order_by: Column to order the elements.
        :param order_direction: Direction to order the elements, either `asc` or `desc`.
        :param filters: Filters to filter the elements.
        :return: Async iterable of the elements.
        """
        return self._table_manager.list(*clauses, order_by=order_by, order_direction=order_direction, **filters)

    async def drop(self, *clauses, **filters) -> int:
        """Drops elements in the repository.

        Returns the number of elements dropped. If no clauses or filters are given, it deletes all the elements in the
        repository.

        Clauses are used to filter the elements using sqlalchemy clauses. Filters are used to filter the elements using
        exact values to specific columns. Clauses and filters can be combined.

        Clause example: `table.c["id"].in_((1, 2, 3))`
        Filter example: `id=1`

        :param clauses: Clauses to filter the elements.
        :param filters: Filters to filter the elements.
        :return: The number of elements dropped.
        """
        return await self._table_manager.drop(*clauses, **filters)



import functools
import inspect
import typing as t

from flama import http, schemas

__all__ = ["PageNumberMixin", "PageNumberResponse"]

from flama.pagination.decorators import PaginationDecoratorFactory


class PageNumberResponse(http.APIResponse):
    """
    Response paginated based on a page number and a page size.

    First 10 elements:
        /resource?page=1
    Third 10 elements:
        /resource?page=3
    First 20 elements:
        /resource?page=1&page_size=20
    """

    default_page_size = 10

    def __init__(
        self,
        schema: t.Any,
        page: t.Optional[t.Union[int, str]] = None,
        page_size: t.Optional[t.Union[int, str]] = None,
        count: t.Optional[bool] = True,
        **kwargs,
    ):
        self.page_number = int(page) if page is not None else 1
        self.page_size = int(page_size) if page_size is not None else self.default_page_size
        self.count = count
        super().__init__(schema=schema, **kwargs)

    def render(self, content: t.Sequence[t.Any]):
        init = (self.page_number - 1) * self.page_size
        end = self.page_number * self.page_size

        return super().render(
            {
                "meta": {
                    "page": self.page_number,
                    "page_size": self.page_size,
                    "count": len(content) if self.count else None,
                },
                "data": content[init:end],
            }
        )


class PageNumberDecoratorFactory(PaginationDecoratorFactory):
    PARAMETERS = [
        inspect.Parameter(
            name="page", default=None, annotation=t.Optional[int], kind=inspect.Parameter.POSITIONAL_OR_KEYWORD
        ),
        inspect.Parameter(
            name="page_size", default=None, annotation=t.Optional[int], kind=inspect.Parameter.POSITIONAL_OR_KEYWORD
        ),
        inspect.Parameter(
            name="count", default=False, annotation=t.Optional[bool], kind=inspect.Parameter.POSITIONAL_OR_KEYWORD
        ),
    ]

    @classmethod
    def _decorate_async(cls, func: t.Callable, schema: t.Any) -> t.Callable:
        @functools.wraps(func)
        async def decorator(
            *args,
            page: t.Optional[int] = None,
            page_size: t.Optional[int] = None,
            count: t.Optional[bool] = False,
            **kwargs,
        ):
            return PageNumberResponse(
                schema=schema, page=page, page_size=page_size, count=count, content=await func(*args, **kwargs)
            )

        return decorator

    @classmethod
    def _decorate_sync(cls, func: t.Callable, schema: t.Any) -> t.Callable:
        @functools.wraps(func)
        def decorator(
            *args,
            page: t.Optional[int] = None,
            page_size: t.Optional[int] = None,
            count: t.Optional[bool] = False,
            **kwargs,
        ):
            return PageNumberResponse(
                schema=schema, page=page, page_size=page_size, count=count, content=func(*args, **kwargs)
            )

        return decorator


class PageNumberMixin:
    def _paginate_page_number(self, func: t.Callable) -> t.Callable:
        """
        Decorator for adding pagination behavior to a view. That decorator produces a view based on page numbering and
        it adds three query parameters to control the pagination: page, page_size and count. Page has a default value of
        first page, page_size default value is defined in
        :class:`PageNumberResponse` and count defines if the response will define
        the total number of elements.

        The output field is also modified by :class:`PageNumberSchema`, creating
        a new field based on it but using the old output field as the content of its data field.

        :param schema_name: Name used for output field.
        :return: Decorated view.
        """
        schema_wrapped = schemas.Schema.from_type(inspect.signature(func).return_annotation)
        resource_schema = schema_wrapped.unique_schema
        schema_name = schema_wrapped.name

        try:
            schema_module, schema_class = schema_name.rsplit(".", 1)
            paginated_schema_name = f"{schema_module}.PageNumberPaginated{schema_class}"
        except ValueError:  # pragma: no cover
            paginated_schema_name = f"PageNumberPaginated{schema_name}"
        schema = schemas.Schema.build(
            paginated_schema_name,
            schema=schemas.schemas.PageNumber,
            fields=[schemas.Field("data", resource_schema, multiple=True)],
        ).unique_schema

        decorator = PageNumberDecoratorFactory.decorate(func, schema)

        self.schemas.update({schema_name: resource_schema, paginated_schema_name: schema})  # type: ignore[attr-defined]

        return decorator



import re
import typing as t

from flama.resources import data_structures
from flama.resources.exceptions import ResourceAttributeError

__all__ = ["Resource", "ResourceType"]


class ResourceType(type):
    METHODS: t.Sequence[str] = ()

    def __new__(mcs, name: str, bases: tuple[type], namespace: dict[str, t.Any]):
        """Resource metaclass for defining basic behavior:
        * Create _meta attribute containing some metadata (name, verbose name...).
        * Adds methods related to resource (create, retrieve...) listed in METHODS class attribute.
        * Generate a Router with above methods.

        :param name: Class name.
        :param bases: List of superclasses.
        :param namespace: Variables namespace used to create the class.
        """
        if not mcs._is_abstract(namespace):
            try:
                # Define resource names
                resource_name, verbose_name = mcs._get_resource_name(name, namespace)
            except AttributeError as e:
                raise ResourceAttributeError(str(e), name)

            namespace.setdefault("_meta", data_structures.Metadata()).name = resource_name
            namespace["_meta"].verbose_name = verbose_name

            # Create methods and routes
            namespace.update(mcs._build_methods(namespace))
            namespace["routes"] = mcs._build_routes(namespace)

        return super().__new__(mcs, name, bases, namespace)

    @staticmethod
    def _is_abstract(namespace: dict[str, t.Any]) -> bool:
        return namespace.get("__module__") == "flama.resources.resource" and namespace.get("__qualname__") == "Resource"

    @classmethod
    def _get_mro(cls, *classes: type) -> list[type]:
        """Generate the MRO list for given base class or list of base classes.

        :param classes: Base classes.
        :return: MRO list.
        """
        return list(
            dict.fromkeys([y for x in [[c.__mro__[0]] + cls._get_mro(*c.__mro__[1:]) for c in classes] for y in x])
        )

    @classmethod
    def _get_attribute(
        cls,
        attribute: str,
        bases: t.Sequence[t.Any],
        namespace: dict[str, t.Any],
        metadata_namespace: t.Optional[str] = None,
    ) -> t.Any:
        """Look for an attribute given his name on namespace or parent classes namespace.

        :param attribute: Attribute name.
        :param bases: List of superclasses.
        :param namespace: Variables namespace used to create the class.
        :return: Attribute.
        """
        try:
            return namespace.pop(attribute)
        except KeyError:
            for base in cls._get_mro(*bases):
                if hasattr(base, "_meta"):
                    if attribute in base._meta.namespaces.get(metadata_namespace, {}):
                        return base._meta.namespaces[metadata_namespace][attribute]

                    if hasattr(base._meta, attribute):
                        return getattr(base._meta, attribute)

                if hasattr(base, attribute):
                    return getattr(base, attribute)

        raise AttributeError(ResourceAttributeError.ATTRIBUTE_NOT_FOUND.format(attribute=attribute))

    @classmethod
    def _get_resource_name(cls, name: str, namespace: dict[str, t.Any]) -> tuple[str, str]:
        """Look for a resource name in namespace and check it's a valid name.

        :param name: Class name.
        :param namespace: Variables namespace used to create the class.
        :return: Resource name.
        """
        resource_name = namespace.pop("name", name)

        # Check resource name validity
        if re.match("[a-zA-Z][-_a-zA-Z]", resource_name) is None:
            raise AttributeError(ResourceAttributeError.RESOURCE_NAME_INVALID.format(resource_name=resource_name))

        return resource_name, namespace.pop("verbose_name", resource_name)

    @classmethod
    def _build_routes(cls, namespace: dict[str, t.Any]) -> dict[str, t.Callable]:
        """Builds the routes' descriptor.

        :param namespace: Variables namespace used to create the class.
        """
        return {
            name: m
            for name, m in namespace.items()
            if hasattr(m, "_meta") and isinstance(m._meta, data_structures.MethodMetadata) and not name.startswith("_")
        }

    @classmethod
    def _build_methods(cls, namespace: dict[str, t.Any]) -> dict[str, t.Callable]:
        """Builds a namespace containing all resource methods. Look for all methods listed in METHODS attribute and
        named '_add_[method]'.

        :param namespace: Variables namespace used to create the class.
        :return: Methods namespace.
        """
        # Get available methods
        methods = [getattr(cls, f"_add_{method}") for method in cls.METHODS if hasattr(cls, f"_add_{method}")]

        # Generate methods
        methods_namespace = {
            func_name: func
            for method in methods
            for func_name, func in method(**namespace["_meta"].to_plain_dict()).items()
        }

        # Preserve already defined methods
        methods_namespace.update(
            {method: methods_namespace[f"_{method}"] for method in cls.METHODS if method not in namespace}
        )

        return methods_namespace


class Resource(metaclass=ResourceType):
    name: str
    verbose_name: str
    _meta: data_structures.Metadata



import typing as t

from flama import exceptions
from flama.models.base import Model

try:
    import torch  # type: ignore
except Exception:  # pragma: no cover
    torch = None  # type: ignore


class PyTorchModel(Model):
    def predict(self, x: list[list[t.Any]]) -> t.Any:
        if torch is None:  # noqa
            raise exceptions.FrameworkNotInstalled("pytorch")

        try:
            return self.model(torch.Tensor(x)).tolist()
        except ValueError as e:
            raise exceptions.HTTPException(status_code=400, detail=str(e))



import abc
import dataclasses
import re
import typing as t
import urllib.parse
import uuid
from decimal import Decimal

T = t.TypeVar("T", bound=t.Union[int, str, float, Decimal, uuid.UUID])


@dataclasses.dataclass
class URL:
    scheme: str
    netloc: str
    path: str
    params: str
    query: str
    fragment: str

    def __init__(self, url: str = "", **kwargs: str):
        """URL object.

        :param url: URL string to be parsed.
        :param kwargs: Individual components to replace those parsed from string.
        """
        parsed_url: urllib.parse.ParseResult = urllib.parse.urlparse(url)._replace(**kwargs)
        self.scheme = parsed_url.scheme
        self.netloc = parsed_url.netloc
        self.path = parsed_url.path
        self.params = parsed_url.params
        self.query = parsed_url.query
        self.fragment = parsed_url.fragment

    @property
    def components(self) -> dict[str, t.Optional[str]]:
        """URL components map.

        :return: Components.
        """
        return dataclasses.asdict(self)

    @property
    def url(self) -> str:
        """Build URL string.

        :return: URL string.
        """
        return str(urllib.parse.urlunparse(tuple(self.components.values())))

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return f"URL('{self.url}')"


class ParamSerializer(t.Generic[T], metaclass=abc.ABCMeta):
    regex: t.ClassVar[str] = ""

    @abc.abstractmethod
    def load(self, value: str) -> T:
        ...

    @abc.abstractmethod
    def dump(self, value: T) -> str:
        ...

    def __eq__(self, other):
        return type(other) == type(self)


class StringParamSerializer(ParamSerializer[str]):
    regex = "[^/]+"

    def load(self, value: str) -> str:
        return str(value)

    def dump(self, value: str) -> str:
        return str(value)


class PathParamSerializer(StringParamSerializer):
    regex = ".*"


class IntegerParamSerializer(ParamSerializer[int]):
    regex = "-?[0-9]+"

    def load(self, value: str) -> int:
        return int(value)

    def dump(self, value: int) -> str:
        return str(value)


class FloatParamSerializer(ParamSerializer[float]):
    regex = "-?[0-9]+(.[0-9]+)?"

    def load(self, value: str) -> float:
        return float(value)

    def dump(self, value: float) -> str:
        return f"{value:0.10f}".rstrip("0").rstrip(".")


class DecimalParamSerializer(ParamSerializer[Decimal]):
    regex = "-?[0-9]+(.[0-9]+)?"

    def load(self, value: str) -> Decimal:
        return Decimal(value)

    def dump(self, value: Decimal) -> str:
        return str(value)


class UUIDParamSerializer(ParamSerializer[uuid.UUID]):
    regex = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"

    def load(self, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    def dump(self, value: uuid.UUID) -> str:
        return str(value)


class RegexPath:
    PARAM_REGEX: t.ClassVar[re.Pattern] = re.compile(
        r"{(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)(?::(?P<type>[a-zA-Z_][a-zA-Z0-9_]*)?)?}"
    )
    SERIALIZERS: t.ClassVar[dict[str, ParamSerializer]] = {
        "str": StringParamSerializer(),
        "path": PathParamSerializer(),
        "int": IntegerParamSerializer(),
        "float": FloatParamSerializer(),
        "decimal": DecimalParamSerializer(),
        "uuid": UUIDParamSerializer(),
    }

    def __init__(self, path: t.Union[str, "RegexPath"]):
        """URL path with a regex to allow path params as placeholders.

        Given a path string like: "/foo/{bar:str}"
        path:        "/foo/{bar:str}"
        template:    "/foo/{bar}"
        regex:       "^/foo/(?P<bar>[^/]+)$"
        serializers: {"bar": StringParamSerializer()}

        Path params is a special param defined for nesting routes, it will be removed from path and template, so for a
        path like: "/foo{path:path}" the result will be:
        path:        "/foo"
        template:    "/foo"
        regex:       "^/foo(?P<path>.*)$"
        serializers: {"path": PathParamSerializer()}

        :param path: URL path.
        """
        if isinstance(path, RegexPath):
            self.raw_path: str = path.raw_path
            self.path: str = path.path
            self.template: str = path.template
            self.regex: re.Pattern = path.regex
            self.serializers: dict[str, ParamSerializer] = path.serializers
            self.parameters: list[str] = path.parameters
        else:
            self.raw_path = path
            self.path = self.PARAM_REGEX.sub(lambda x: x.group(0) if x.group("type") != "path" else "", path)

            assert self.path == "" or self.path.startswith("/"), "Routed paths must start with '/'"

            self.template = self.PARAM_REGEX.sub(
                lambda x: f"{{{x.group('name')}}}" if x.group("type") != "path" else "", path
            )
            regex = self.PARAM_REGEX.sub(
                lambda x: f"(?P<{x.group('name')}>{self._serializer(x.group('type')).regex})", path
            )
            self.regex = re.compile(rf"^{regex}$")
            self.serializers = {
                param_name: self._serializer(param_type) for param_name, param_type in self.PARAM_REGEX.findall(path)
            }
            self.parameters = list(self.serializers.keys())

    def _serializer(self, param_type: t.Optional[str]) -> ParamSerializer:
        try:
            return self.SERIALIZERS[param_type or "str"]
        except KeyError:
            raise ValueError(f"Unknown path param serializer '{param_type}'")

    def match(self, path: str) -> bool:
        """Check if given path matches with current object.

        :param path: Path to match
        :return: True if matches.
        """
        return self.regex.match(path) is not None

    def values(self, path: str) -> dict[str, t.Any]:
        """Get serialized parameters from a matching path.

        :param path: Path to match.
        :return: Path param values.
        """
        match = self.regex.match(path)

        if match is None:
            raise ValueError(f"Path '{path}' does not match.")

        return {k: self.serializers[k].load(v) for k, v in match.groupdict().items()}

    def build(self, **params: t.Any) -> tuple[str, dict[str, t.Any]]:
        """Build a path by completing param placeholders with given values.

        :param params: Param values.
        :return: Built path and unused params.
        """
        if not set(self.serializers.keys()) <= set(params.keys()):
            formatted_params = ", ".join(f"'{x}'" for x in self.serializers.keys())
            raise ValueError(f"Wrong params, expected: {formatted_params}.")

        if not self.serializers:
            return self.template, params

        remaining_params = {k: v for k, v in params.items() if k not in self.serializers}
        values = {k: self.serializers[k].dump(v) for k, v in params.items() if k in self.serializers}
        path = re.sub(
            pattern=f"{{({r'|'.join(values.keys())})}}",
            repl=lambda x: values.pop(x.group(1)),
            string=self.template,
        )
        return path, {**values, **remaining_params}

    def __eq__(self, other) -> bool:
        return isinstance(other, RegexPath) and self.path.__eq__(other.path) or self.path.__eq__(other)

    def __str__(self) -> str:
        return self.path.__str__()

    def __repr__(self) -> str:
        return self.path.__repr__()

    def __add__(self, other) -> "RegexPath":
        if isinstance(other, RegexPath):
            return RegexPath(self.path + other.path)
        elif isinstance(other, str):
            return RegexPath(self.path + other)

        raise TypeError("Can only concatenate str or Path to Path")



import inspect
import typing as t

from flama import schemas
from flama.injection.resolver import Return
from flama.schemas.data_structures import Field, Parameter, Parameters

if t.TYPE_CHECKING:
    from flama.applications import Flama
    from flama.injection.resolver import Parameter as InjectionParameter
    from flama.routing import BaseRoute

__all__ = ["RouteParametersMixin"]


class ParametersDescriptor:
    def __init__(self) -> None:
        self._route: "BaseRoute"
        self._app: "Flama"

    def __get__(self, instance, owner) -> "ParametersDescriptor":
        self._route = instance
        return self

    @property
    def _parameters(self) -> dict[str, list["InjectionParameter"]]:
        return {
            method: sorted(
                [
                    parameter
                    for resolution in self._app.injector.resolve_function(handler).values()
                    for parameter in resolution.parameters
                ],
                key=lambda x: x.name,
            )
            for method, handler in self._route.endpoint_handlers().items()
        }

    @property
    def _return_values(self) -> dict[str, "InjectionParameter"]:
        return {
            method: Return.from_return_annotation(inspect.signature(handler).return_annotation)
            for method, handler in self._route.endpoint_handlers().items()
        }

    @property
    def query(self) -> dict[str, Parameters]:
        return {
            method: {
                p.name: Parameter.build("query", p)
                for p in parameters
                if Field.is_http_valid_type(p.annotation) and p.name not in self._route.path.parameters
            }
            for method, parameters in self._parameters.items()
        }

    @property
    def path(self) -> dict[str, Parameters]:
        return {
            method: {p.name: Parameter.build("path", p) for p in parameters if p.name in self._route.path.parameters}
            for method, parameters in self._parameters.items()
        }

    @property
    def body(self) -> dict[str, t.Optional[Parameter]]:
        return {
            method: next(
                (
                    Parameter.build("body", p)
                    for p in parameters
                    if (schemas.is_schema(p.annotation) or t.get_origin(p.annotation) == list)
                    and p.name not in self._route.path.parameters
                ),
                None,
            )
            for method, parameters in self._parameters.items()
        }

    @property
    def response(self) -> dict[str, Parameter]:
        return {
            method: Parameter.build("response", return_value) for method, return_value in self._return_values.items()
        }

    def build(self, app: "Flama") -> "ParametersDescriptor":
        self._app = app
        return self


class RouteParametersMixin:
    """
    Mixin for adding fields discovery behavior to Routes.
    """

    parameters = ParametersDescriptor()



import json
import typing as t

from flama import exceptions
from flama.codecs.base import WebsocketsCodec

if t.TYPE_CHECKING:
    from flama import types

__all__ = ["JSONCodec"]


class JSONCodec(WebsocketsCodec):
    encoding = "json"

    async def decode(self, item: "types.Message", **options):
        if item.get("text") is not None:
            text = item["text"]
        else:
            text = item["bytes"].decode("utf-8")

        try:
            return json.loads(text)
        except json.decoder.JSONDecodeError:
            raise exceptions.DecodeError("Malformed JSON data received")



import abc
import typing as t
from collections import defaultdict

if t.TYPE_CHECKING:
    from flama import Flama

__all__ = ["Module", "Modules"]


class _BaseModule:
    name: str

    def __init__(self) -> None:
        self.app: "Flama"

    async def on_startup(self) -> None:
        ...

    async def on_shutdown(self) -> None:
        ...


class _ModuleMeta(abc.ABCMeta):
    def __new__(mcs, name, bases, namespace):
        if _BaseModule not in bases:
            assert namespace.get("name"), f"Module '{name}' does not have a 'name' attribute."
        return super().__new__(mcs, name, bases, namespace)


class Module(_BaseModule, metaclass=_ModuleMeta):
    ...


class Modules(dict[str, Module]):
    def __init__(self, app: "Flama", modules: t.Optional[t.Union[t.Sequence[Module], set[Module]]]):
        modules_map: dict[str, list[Module]] = defaultdict(list)
        for module in modules or []:
            module.app = app
            modules_map[module.name].append(module)

        collisions = {name: {x.__class__.__name__ for x in mods} for name, mods in modules_map.items() if len(mods) > 1}
        assert not collisions, "Collision in module names: " + ", ".join(
            f"{name} ({', '.join(sorted(mods))})" for name, mods in collisions.items()
        )

        super().__init__({name: mods[0] for name, mods in modules_map.items()})

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (list, tuple, set)):
            return {module.__class__ for module in self.values()} == set(other)  # type: ignore

        return super().__eq__(other)



__all__ = [
    "AuthenticationException",
    "Unauthorized",
    "JWTException",
    "JWTDecodeException",
    "JWTValidateException",
    "JWTClaimValidateException",
]


class AuthenticationException(Exception):
    ...


class Unauthorized(AuthenticationException):
    ...


class Forbidden(AuthenticationException):
    ...


class JWTException(AuthenticationException):
    ...


class JWTDecodeException(JWTException):
    ...


class JWTValidateException(JWTException):
    ...


class JWTClaimValidateException(JWTValidateException):
    def __init__(self, claim: str) -> None:
        self.claim = claim
        super().__init__(f"Claim '{self.claim}' is not valid")



import dataclasses
import functools
import json
import logging
import os
import typing as t

from flama.config import exceptions, types
from flama.config.data_structures import FileDict

__all__ = ["Config"]

logger = logging.getLogger(__name__)

R = t.TypeVar("R")
UNKNOWN = types.UnknownType("")


class Config:
    """Tool for retrieving config parameters from a config file or environment variables.

    This class can be used to retrieve config parameters only from environment variables:
        >>> config = Config()

    But it can handle several config file formats, some examples:
        >>> config_ini = Config(".env", "ini")
        >>> config_json = Config("config.json", "json")
        >>> config_yaml = Config("config.yaml", "yaml")
        >>> config_toml = Config("config.toml", "toml")

    Once the config object is created it allows to get config parameters using following syntax:
        >>> FOO = config("FOO", default="bar")

    The value of the config parameter will be looked for based on the following order:
    1. Environment variable
    2. Config file
    3. Explicit default value

    It is possible to convert config parameters into python types by specifying the type as part of the call. In this
    case an environment variable contains the values 'true' or 'false':
        >>> DEBUG = config("DEBUG", cast=bool)

    Also, a more complex conversion is possible by using custom functions, lets take the example of a variable where
    zero means False and any other value than zero means a True:
        >>> DEBUG = config("DEBUG", cast=lambda x: x != 0)

    For last, there is a case when the config parameter is a json valid value, in this case it is possible to convert
    it into a dataclass:
        >>> @dataclasses.dataclass
        ... class Puppy:
        ...     name: str
        ...     age: int
        >>> PUPPY = config("PUPPY", cast=Puppy)
    """

    def __init__(
        self,
        config_file: t.Optional[t.Union[str, os.PathLike]] = None,
        format: t.Union[str, types.FileFormat] = types.FileFormat.ini,
    ) -> None:
        """Tool for retrieving config parameters from a config file or environment variables.

        :param config_file: Config file path.
        :param format: Config file format.
        """
        if config_file:
            try:
                self.config_file = FileDict(config_file, format)
                logger.info("Config file '%s' loaded", config_file)
            except exceptions.LoadError:
                logger.info("Config file '%s' cannot be loaded", config_file)
                self.config_file = {}
        else:
            self.config_file = {}

    def _get_item_from_environment(self, key: str) -> t.Any:
        return os.environ[key]

    def _get_item_from_config_file(self, key: str) -> t.Any:
        return functools.reduce(lambda x, k: x[k], key.split("."), self.config_file)

    def _get_item(self, key: str, default: t.Union[R, types.UnknownType] = UNKNOWN) -> R:
        try:
            return self._get_item_from_environment(key)
        except KeyError:
            ...

        try:
            return self._get_item_from_config_file(key)
        except KeyError:
            ...

        if default is not UNKNOWN:
            return t.cast(R, default)

        raise KeyError(key)

    def _build_dataclass(self, data: t.Any, dataclass: type[R]) -> R:
        if isinstance(data, str):
            try:
                data = json.loads(data)
            except Exception as e:
                raise exceptions.ConfigError("Cannot parse value as json for config dataclass") from e

        if not isinstance(data, dict):
            raise exceptions.ConfigError("Wrong value for config dataclass")

        try:
            fields = [f.name for f in dataclasses.fields(dataclass)]  # type: ignore
            return dataclass(**{k: v for k, v in data.items() if k in fields})
        except Exception as e:
            raise exceptions.ConfigError("Cannot create config dataclass") from e

    @t.overload
    def __call__(self, key: str) -> t.Any:
        ...

    @t.overload
    def __call__(self, key: str, *, default: t.Union[R, types.UnknownType]) -> R:
        ...

    @t.overload
    def __call__(self, key: str, *, cast: type[R]) -> R:
        ...

    @t.overload
    def __call__(self, key: str, *, default: t.Union[R, types.UnknownType], cast: type[R]) -> R:
        ...

    @t.overload
    def __call__(self, key: str, *, cast: t.Callable[[t.Any], R]) -> R:
        ...

    @t.overload
    def __call__(self, key: str, *, default: t.Union[R, types.UnknownType], cast: t.Callable[[t.Any], R]) -> R:
        ...

    def __call__(
        self,
        key: str,
        *,
        default: t.Union[R, types.UnknownType] = UNKNOWN,
        cast: t.Optional[t.Union[type[R], t.Callable[[t.Any], R]]] = None,
    ) -> R:
        """Get config parameter value.

        :param key: Config parameter name.
        :param default: Default value if config parameter is not found.
        :param cast: Type or function to convert config parameter value.
        :return: Config parameter value.
        :raises KeyError: If config parameter is not found and no default value is specified.
        :raises ConfigError: If config parameter cannot be converted to the specified type.

        Examples:
            >>> config = Config(".env", "ini")
            >>> FOO = config("FOO", default="bar")  # Default value if FOO is not found
            >>> DEBUG = config("DEBUG", cast=bool)  # Convert value to boolean
            >>> @dataclasses.dataclass
            ... class Puppy:
            ...     name: str
            ...     age: int
            >>> PUPPY = config("PUPPY", cast=Puppy)  # Parse json value and convert it to a dataclass
        """
        value = self._get_item(key, default)

        if cast is None:
            return value

        if dataclasses.is_dataclass(cast) and isinstance(cast, type):
            return t.cast(R, self._build_dataclass(data=value, dataclass=cast))

        try:
            return t.cast(t.Callable[[t.Any], R], cast)(value)
        except Exception as e:
            raise exceptions.ConfigError("Cannot cast config type") from e



import inspect
import itertools
import typing as t
import warnings

import typesystem

from flama import compat
from flama.injection import Parameter
from flama.schemas._libs.typesystem.fields import MAPPING, MAPPING_TYPES
from flama.schemas.adapter import Adapter
from flama.schemas.exceptions import SchemaGenerationError, SchemaValidationError
from flama.types import JSONSchema

__all__ = ["TypesystemAdapter"]

Schema = typesystem.Schema
Field = typesystem.Field


class TypesystemAdapter(Adapter[Schema, Field]):
    def build_field(
        self,
        name: str,
        type_: type,
        nullable: bool = False,
        required: bool = True,
        default: t.Any = None,
        multiple: bool = False,
        **kwargs,
    ) -> Field:
        if required is False and default is not Parameter.empty:
            kwargs["default"] = default

        kwargs.update({"allow_null": nullable, "title": name})

        if multiple:
            return typesystem.Array(
                (
                    typesystem.Reference(to=type_.title, definitions=typesystem.Definitions({type_.title: type_}))
                    if self.is_schema(type_)
                    else MAPPING[type_]()
                ),
                **kwargs,
            )

        return MAPPING[type_](**kwargs)

    def build_schema(  # type: ignore[return-value]
        self,
        *,
        name: t.Optional[str] = None,
        schema: t.Optional[t.Union[Schema, type[Schema]]] = None,
        fields: t.Optional[dict[str, Field]] = None,
        partial: bool = False,
    ) -> Schema:
        fields_ = {**(self.unique_schema(schema).fields if self.is_schema(schema) else {}), **(fields or {})}

        if partial:
            for field in fields_:
                fields_[field].default = None
                fields_[field].allow_null = True

        return Schema(title=name or self.DEFAULT_SCHEMA_NAME, fields=fields_)

    def validate(self, schema: Schema, values: dict[str, t.Any], *, partial: bool = False) -> t.Any:
        try:
            if partial:
                warnings.warn("Typesystem does not support partial validation")
                return schema.serialize(values)
            return schema.validate(values)
        except typesystem.ValidationError as errors:
            raise SchemaValidationError(errors={k: [v] for k, v in errors.items()})

    def load(self, schema: Schema, value: dict[str, t.Any]) -> t.Any:
        return schema.validate(value)

    def dump(self, schema: Schema, value: dict[str, t.Any]) -> t.Any:
        return self._dump(self.validate(schema, value))

    def _dump(self, value: t.Any) -> t.Any:
        if isinstance(value, list):
            return [self._dump(x) for x in value]

        return value

    def name(self, schema: Schema, *, prefix: t.Optional[str] = None) -> str:
        if not schema.title:
            raise ValueError(f"Schema '{schema}' needs to define title attribute")

        schema_name = f"{prefix or ''}{schema.title}"
        return schema_name if schema.__module__ == "builtins" else f"{schema.__module__}.{schema_name}"

    def to_json_schema(self, schema: t.Union[Schema, Field]) -> JSONSchema:
        try:
            json_schema = typesystem.to_json_schema(schema)

            if not isinstance(json_schema, dict):
                raise SchemaGenerationError

            for property in itertools.chain(json_schema.get("properties", {}).values(), [json_schema]):
                if isinstance(property.get("type"), list):
                    property["anyOf"] = [{"type": x} for x in property["type"]]
                    del property["type"]

            json_schema.pop("components", None)
        except Exception as e:
            raise SchemaGenerationError from e

        return json_schema

    def unique_schema(self, schema: Schema) -> Schema:
        return schema

    def _get_field_type(self, field: Field) -> t.Any:
        if isinstance(field, typesystem.Reference):
            return field.target

        if isinstance(field, typesystem.Array):
            return (
                [self._get_field_type(x) for x in field.items]
                if isinstance(field.items, (list, tuple, set))
                else self._get_field_type(field.items)
            )

        if isinstance(field, typesystem.Object):
            object_fields = {k: self._get_field_type(v) for k, v in field.properties.items()}
            if isinstance(field.additional_properties, (typesystem.Field, typesystem.Reference)):
                object_fields[""] = self._get_field_type(field.additional_properties)
            return object_fields

        try:
            return MAPPING_TYPES[field.__class__]
        except KeyError:
            return None

    def schema_fields(
        self, schema: Schema
    ) -> dict[
        str,
        tuple[
            t.Union[t.Union[Schema, type], list[t.Union[Schema, type]], dict[str, t.Union[Schema, type]]],
            Field,
        ],
    ]:
        return {name: (self._get_field_type(field), field) for name, field in schema.fields.items()}

    def is_schema(self, obj: t.Any) -> compat.TypeGuard[Schema]:  # PORT: Replace compat when stop supporting 3.9
        return isinstance(obj, Schema) or (inspect.isclass(obj) and issubclass(obj, Schema))

    def is_field(self, obj: t.Any) -> compat.TypeGuard[Field]:  # PORT: Replace compat when stop supporting 3.9
        return isinstance(obj, Field) or (inspect.isclass(obj) and issubclass(obj, Field))



__all__ = ["ConfigError", "LoadError"]


class ConfigError(Exception):
    """Exception raised when there is a problem with the config."""

    ...


class LoadError(ConfigError):
    """Exception raised when there is a problem with the config loading."""

    ...



import enum
import typing as t

from flama import compat

__all__ = ["FileFormat"]

UnknownType = t.NewType("UnknownType", str)


class FileFormat(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
    """Config file format."""

    ini = enum.auto()
    json = enum.auto()
    yaml = enum.auto()
    toml = enum.auto()



import asyncio
import logging
import typing as t

from flama import concurrency, exceptions, types

if t.TYPE_CHECKING:
    from flama import Flama

__all__ = ["Lifespan"]

logger = logging.getLogger(__name__)


class Lifespan(types.AppClass):
    def __init__(self, lifespan: t.Optional[t.Callable[[t.Optional["Flama"]], t.AsyncContextManager]] = None):
        """A class that handles the lifespan of an application.

        It is responsible for calling the startup and shutdown events and the user defined lifespan.

        :param lifespan: A user defined lifespan. It must be a callable that returns an async context manager.
        """
        self.lifespan = lifespan
        self.lock = asyncio.Lock()

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Handles a lifespan request by initialising and finalising all modules and running a user defined lifespan.

        :param scope: ASGI request.
        :param receive: ASGI receive.
        :param send: ASGI send.
        """
        async with self.lock:
            app = scope["app"]
            message = await receive()
            logger.debug("Start lifespan for app '%s' from status '%s' with message '%s'", app, app.status, message)
            if message["type"] == "lifespan.startup":
                if app.status not in (types.AppStatus.NOT_STARTED, types.AppStatus.SHUT_DOWN):
                    msg = f"Trying to start application from '{app._status}' state"
                    await send(types.Message({"type": "lifespan.startup.failed", "message": msg}))
                    raise exceptions.ApplicationError(msg)

                try:
                    logger.info("Application starting")
                    app.status = types.AppStatus.STARTING
                    await self._startup(app)
                    await self._child_propagation(app, scope, message)
                    app.status = types.AppStatus.READY
                    await send(types.Message({"type": "lifespan.startup.complete"}))
                    logger.info("Application ready")
                except BaseException as e:
                    logger.exception("Application start failed")
                    app.status = types.AppStatus.FAILED
                    await send(types.Message({"type": "lifespan.startup.failed", "message": str(e)}))
                    raise exceptions.ApplicationError("Lifespan startup failed") from e
            elif message["type"] == "lifespan.shutdown":
                if app.status != types.AppStatus.READY:
                    msg = f"Trying to shutdown application from '{app._status}' state"
                    await send(types.Message({"type": "lifespan.shutdown.failed", "message": msg}))
                    raise exceptions.ApplicationError(msg)

                try:
                    logger.info("Application shutting down")
                    app.status = types.AppStatus.SHUTTING_DOWN
                    await self._child_propagation(app, scope, message)
                    await self._shutdown(app)
                    app.status = types.AppStatus.SHUT_DOWN
                    await send(types.Message({"type": "lifespan.shutdown.complete"}))
                    logger.info("Application shut down")
                except BaseException as e:
                    await send(types.Message({"type": "lifespan.shutdown.failed", "message": str(e)}))
                    app.status = types.AppStatus.FAILED
                    logger.exception("Application shutdown failed")
                    raise exceptions.ApplicationError("Lifespan shutdown failed") from e
            else:
                logger.warning("Unknown lifespan message received: %s", str(message))

            logger.debug("End lifespan for app '%s' with status '%s'", app, app.status)

    async def _startup(self, app: "Flama") -> None:
        if app.events.startup:
            await concurrency.run_task_group(*(f() for f in app.events.startup))

        if self.lifespan:
            await self.lifespan(app).__aenter__()

    async def _shutdown(self, app: "Flama") -> None:
        if self.lifespan:
            await self.lifespan(app).__aexit__(None, None, None)

        if app.events.shutdown:
            await concurrency.run_task_group(*(f() for f in app.events.shutdown))

    async def _child_propagation(self, app: "Flama", scope: types.Scope, message: types.Message) -> None:
        async def child_receive() -> types.Message:
            return message

        async def child_send(message: types.Message) -> None:
            ...

        if app.routes:
            await concurrency.run_task_group(*(route(scope, child_receive, child_send) for route in app.routes))



import inspect
import typing as t

from flama.pagination import paginator
from flama.resources import data_structures
from flama.routing import Mount, Route

if t.TYPE_CHECKING:
    from flama import Flama
    from flama.pagination.types import PaginationType
    from flama.resources import Resource

__all__ = ["ResourceRoute", "resource_method"]


class ResourceRoute(Mount):
    def __init__(
        self,
        path: str,
        resource: t.Union["Resource", type["Resource"]],
        tags: t.Optional[dict[str, dict[str, t.Any]]] = None,
    ):
        tags = tags or {}

        # Handle class or instance objects
        self.resource = resource() if inspect.isclass(resource) else resource

        assert set(self.resource.routes.keys()) >= set(  # type: ignore
            tags.keys()
        ), "Tags must be defined only for existing routes."

        routes = [
            Route(
                path=route._meta.path,
                endpoint=getattr(self.resource, name),
                methods=route._meta.methods,
                name=route._meta.name or route.__name__,
                tags=tags.get(name, route._meta.tags),
            )
            for name, route in self.resource.routes.items()  # type: ignore
        ]

        super().__init__(path=path, routes=routes, name=self.resource._meta.name)  # type: ignore

    def build(self, app: t.Optional["Flama"] = None) -> None:
        """Build step for resource routes.

        Add this resource's repository to Flama's worker.

        :param app: Flama app.
        """
        from flama import Flama

        super().build(app)

        if (root := (self.app if isinstance(self.app, Flama) else app)) and "ddd" in self.resource._meta.namespaces:
            root.resources.add_repository(
                name=self.resource._meta.name,
                repository=self.resource._meta.namespaces["ddd"]["repository"],
            )


def resource_method(
    path: str,
    methods: t.Optional[t.Sequence[str]] = None,
    name: t.Optional[str] = None,
    *,
    pagination: t.Optional[t.Union[str, "PaginationType"]] = None,
    tags: t.Optional[dict[str, t.Any]] = None,
) -> t.Callable:
    """Decorator for adding useful info needed for generating resource routes.

    :param path: Route path.
    :param methods: HTTP methods available.
    :param name: Route name.
    :param pagination: Apply a pagination technique.
    :param tags: Tags to add to the method.
    :return: Decorated method.
    """

    def wrapper(func: t.Callable) -> t.Callable:
        func = paginator.paginate(pagination, func) if pagination is not None else func

        func._meta = data_structures.MethodMetadata(  # type: ignore
            path=path, methods=set(methods) if methods is not None else {"GET"}, name=name, tags=tags or {}
        )

        return func

    return wrapper



import http
import logging
import typing as t

from flama import authentication
from flama.exceptions import HTTPException
from flama.http import APIErrorResponse, Request

if t.TYPE_CHECKING:
    from flama import Flama, types
    from flama.http import Response
    from flama.routing import BaseRoute

__all__ = ["AuthenticationMiddleware"]


logger = logging.getLogger(__name__)


class AuthenticationMiddleware:
    def __init__(self, app: "types.App"):
        self.app: "Flama" = t.cast("Flama", app)

    async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        response = await self._get_response(scope, receive)

        await response(scope, receive, send)

    def _get_permissions(self, route: "BaseRoute") -> set[str]:
        return set(route.tags.get("permissions", []))

    async def _get_response(self, scope: "types.Scope", receive: "types.Receive") -> t.Union["Response", "Flama"]:
        app: "Flama" = scope["app"]

        route, _ = app.router.resolve_route(scope)

        required_permissions = self._get_permissions(route)

        if not required_permissions:
            return self.app

        try:
            token: authentication.AccessToken = await app.injector.resolve(authentication.AccessToken).value(
                {"request": Request(scope, receive=receive)}
            )
        except HTTPException as e:
            logger.debug("JWT error: %s", e.detail)
            return APIErrorResponse(status_code=e.status_code, detail=e.detail)

        user_permissions = set(token.payload.data.get("permissions", [])) | {
            y for x in token.payload.data.get("roles", {}).values() for y in x
        }
        if not (user_permissions >= required_permissions):
            logger.debug("User does not have the required permissions: %s", required_permissions)
            return APIErrorResponse(status_code=http.HTTPStatus.FORBIDDEN, detail="Insufficient permissions")

        return self.app



import os
import typing as t

from flama.config import exceptions, loaders, types

__all__ = ["FileDict"]


class FileDict(t.Mapping[str, t.Any]):
    """A dictionary that loads its data from a file. Supports json, toml, yaml and ini files."""

    _LOADERS: dict[types.FileFormat, loaders.FileLoader] = {
        types.FileFormat.json: loaders.JSONFileLoader(),
        types.FileFormat.toml: loaders.TOMLFileLoader(),
        types.FileFormat.yaml: loaders.YAMLFileLoader(),
        types.FileFormat.ini: loaders.ConfigFileLoader(),
    }

    def __init__(self, config_file: t.Union[str, os.PathLike], format: t.Union[str, types.FileFormat]):
        """A dictionary that loads its data from a file. Supports json, toml, yaml and ini files.

        :param config_file: Config file path.
        :param format: Config file format.
        """
        try:
            self._loader = self._LOADERS[types.FileFormat[format]]
        except KeyError:
            raise exceptions.ConfigError("Wrong config file format")

        try:
            self._data = self._loader.load(config_file)
        except Exception:
            raise exceptions.LoadError("Config file cannot be loaded")

    def __getitem__(self, key: str) -> t.Any:
        return self._data.__getitem__(key)

    def __eq__(self, other: object) -> bool:
        return self._data.__eq__(other)

    def __iter__(self) -> t.Iterator:
        return self._data.__iter__()

    def __len__(self) -> int:
        return self._data.__len__()

    def __repr__(self) -> str:
        return f"FileDict({self._data.__repr__()})"



import typing as t

from flama import types
from flama.ddd.workers import AbstractWorker
from flama.injection import Component

if t.TYPE_CHECKING:
    from flama.injection import Parameter


__all__ = ["WorkerComponent"]


class WorkerComponent(Component):
    def __init__(self, worker: AbstractWorker):
        self.worker = worker

    def can_handle_parameter(self, parameter: "Parameter") -> bool:
        return parameter.annotation is self.worker.__class__

    def resolve(self, scope: types.Scope):
        self.worker.app = scope["root_app"]
        return self.worker



import typing as t

from flama.pagination import mixins, types

__all__ = ["paginator"]


class Paginator(mixins.LimitOffsetMixin, mixins.PageNumberMixin):
    def __init__(self):
        self.schemas = {}

    def paginate(self, pagination: t.Union[str, types.PaginationType], func: t.Callable) -> t.Callable:
        return {
            types.PaginationType.limit_offset: self._paginate_limit_offset,
            types.PaginationType.page_number: self._paginate_page_number,
        }[types.PaginationType[pagination]](func)


paginator = Paginator()



import typing as t

import click

from flama.cli.config.app import App
from flama.cli.config.app import options as app_options
from flama.cli.config.config import Config
from flama.cli.config.config import options as config_options
from flama.cli.config.uvicorn import Uvicorn
from flama.cli.config.uvicorn import options as uvicorn_options

__all__ = ["serve", "command"]


@click.command(name="serve", context_settings={"auto_envvar_prefix": "FLAMA"})
@config_options
@app_options
@uvicorn_options
def command(app: App, uvicorn: Uvicorn):
    """Serve an ML model file within a Flama Application.

    Serve the ML model file specified by <MODEL_PATH> within a Flama Application."""
    Config(app=app, server=uvicorn).run()


serve: t.Callable = command.callback  # type: ignore[assignment]



import datetime
import os
import typing as t
import uuid

from flama.serialize.data_structures import Compression, ModelArtifact

if t.TYPE_CHECKING:
    from flama.serialize.data_structures import Artifacts

__all__ = ["dump"]


def dump(
    model: t.Any,
    path: t.Union[str, os.PathLike],
    *,
    compression: t.Union[str, Compression] = Compression.standard,
    model_id: t.Optional[t.Union[str, uuid.UUID]] = None,
    timestamp: t.Optional[datetime.datetime] = None,
    params: t.Optional[dict[str, t.Any]] = None,
    metrics: t.Optional[dict[str, t.Any]] = None,
    extra: t.Optional[dict[str, t.Any]] = None,
    artifacts: t.Optional["Artifacts"] = None,
    **kwargs,
) -> None:
    """Serialize an ML model using Flama format to bytes stream.

    :param model: The ML model.
    :param path: Model file path.
    :param compression: Compression type.
    :param model_id: The model ID.
    :param timestamp: The model timestamp.
    :param params: The model parameters.
    :param metrics: The model metrics.
    :param extra: The model extra data.
    :param artifacts: The model artifacts.
    :param kwargs: Keyword arguments passed to library dump method.
    """
    ModelArtifact.from_model(
        model,
        model_id=model_id,
        timestamp=timestamp,
        params=params,
        metrics=metrics,
        extra=extra,
        artifacts=artifacts,
    ).dump(path, compression, **kwargs)



import dataclasses
import typing as t

import flama.types

__all__ = [
    "Schema",
    "Reference",
    "Contact",
    "License",
    "ExternalDocs",
    "Example",
    "Tag",
    "Info",
    "ServerVariable",
    "Server",
    "Link",
    "Security",
    "Callback",
    "Header",
    "Parameter",
    "Encoding",
    "MediaType",
    "RequestBody",
    "Response",
    "Responses",
    "Operation",
    "Path",
    "Paths",
    "Components",
    "OpenAPI",
    "OpenAPISpec",
]

Schema = t.NewType("Schema", flama.types.JSONSchema)


@dataclasses.dataclass(frozen=True)
class Reference:
    ref: str


@dataclasses.dataclass(frozen=True)
class Contact:
    name: t.Optional[str] = None
    url: t.Optional[str] = None
    email: t.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class License:
    name: str
    url: t.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class ExternalDocs:
    url: str
    description: t.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class Example:
    summary: t.Optional[str] = None
    description: t.Optional[str] = None
    value: t.Optional[t.Any] = None
    externalValue: t.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class Tag:
    name: str
    description: t.Optional[str] = None
    externalDocs: t.Optional[ExternalDocs] = None


@dataclasses.dataclass(frozen=True)
class Info:
    title: str
    version: str
    description: t.Optional[str] = None
    termsOfService: t.Optional[str] = None
    contact: t.Optional[Contact] = None
    license: t.Optional[License] = None


@dataclasses.dataclass(frozen=True)
class ServerVariable:
    enum: list[str]
    default: str
    description: t.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class Server:
    url: str
    variables: dict[str, ServerVariable]
    description: t.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class Link:
    operationRef: t.Optional[str] = None
    operationId: t.Optional[str] = None
    parameters: t.Optional[dict[str, t.Any]] = None
    requestBody: t.Optional[t.Any] = None
    description: t.Optional[str] = None
    server: t.Optional[Server] = None


Security = t.NewType("Security", dict[str, list[str]])
Callback = t.NewType("Callback", dict[str, "Path"])


@dataclasses.dataclass(frozen=True)
class Header:
    description: t.Optional[str] = None
    required: t.Optional[bool] = None
    deprecated: t.Optional[bool] = None
    allowEmptyValue: t.Optional[bool] = None
    style: t.Optional[str] = None
    explode: t.Optional[bool] = None
    allowReserved: t.Optional[bool] = None
    schema: t.Optional[t.Union[Schema, Reference]] = None
    example: t.Optional[t.Any] = None
    examples: t.Optional[dict[str, t.Union[Example, Reference]]] = None


@dataclasses.dataclass(frozen=True)
class Parameter:
    name: str
    in_: str
    description: t.Optional[str] = None
    required: t.Optional[bool] = None
    deprecated: t.Optional[bool] = None
    allowEmptyValue: t.Optional[bool] = None
    style: t.Optional[str] = None
    explode: t.Optional[bool] = None
    allowReserved: t.Optional[bool] = None
    schema: t.Optional[t.Union[Schema, Reference]] = None
    example: t.Optional[t.Any] = None
    examples: t.Optional[dict[str, t.Union[Example, Reference]]] = None


@dataclasses.dataclass(frozen=True)
class Encoding:
    contentType: t.Optional[str] = None
    headers: t.Optional[dict[str, t.Union[Header, Reference]]] = None
    style: t.Optional[str] = None
    explode: t.Optional[bool] = None
    allowReserved: t.Optional[bool] = None


@dataclasses.dataclass(frozen=True)
class MediaType:
    schema: t.Optional[t.Union[Schema, Reference]] = None
    example: t.Optional[t.Any] = None
    examples: t.Optional[dict[str, t.Union[t.Any, Reference]]] = None
    encoding: t.Optional[dict[str, Encoding]] = None


@dataclasses.dataclass(frozen=True)
class RequestBody:
    content: dict[str, MediaType]
    description: t.Optional[str] = None
    required: t.Optional[bool] = None


@dataclasses.dataclass(frozen=True)
class Response:
    description: str
    headers: t.Optional[dict[str, t.Union[Header, Reference]]] = None
    content: t.Optional[dict[str, MediaType]] = None
    links: t.Optional[dict[str, t.Union[Link, Reference]]] = None


Responses = t.NewType("Responses", dict[str, Response])


@dataclasses.dataclass(frozen=True)
class Operation:
    responses: Responses
    tags: t.Optional[list[str]] = None
    summary: t.Optional[str] = None
    description: t.Optional[str] = None
    externalDocs: t.Optional[ExternalDocs] = None
    operationId: t.Optional[str] = None
    parameters: t.Optional[list[t.Union[Parameter, Reference]]] = None
    requestBody: t.Optional[t.Union[RequestBody, Reference]] = None
    callbacks: t.Optional[dict[str, t.Union[Callback, Reference]]] = None
    deprecated: t.Optional[bool] = None
    security: t.Optional[list[Security]] = None
    servers: t.Optional[list[Server]] = None


@dataclasses.dataclass(frozen=True)
class Path:
    ref: t.Optional[str] = None
    summary: t.Optional[str] = None
    description: t.Optional[str] = None
    get: t.Optional[Operation] = None
    put: t.Optional[Operation] = None
    post: t.Optional[Operation] = None
    delete: t.Optional[Operation] = None
    options: t.Optional[Operation] = None
    head: t.Optional[Operation] = None
    patch: t.Optional[Operation] = None
    trace: t.Optional[Operation] = None
    servers: t.Optional[list[Server]] = None
    parameters: t.Optional[list[t.Union[Parameter, Reference]]] = None

    @property
    def operations(self) -> dict[str, Operation]:
        return {
            x: getattr(self, x)
            for x in ("get", "put", "post", "delete", "options", "head", "patch", "trace")
            if getattr(self, x) is not None
        }


Paths = t.NewType("Paths", dict[str, Path])


@dataclasses.dataclass(frozen=True)
class Components:
    schemas: dict[str, t.Union[Schema, Reference]]
    responses: dict[str, t.Union[Response, Reference]]
    parameters: dict[str, t.Union[Parameter, Reference]]
    examples: dict[str, t.Union[Example, Reference]]
    requestBodies: dict[str, t.Union[RequestBody, Reference]]
    headers: dict[str, t.Union[Header, Reference]]
    securitySchemes: dict[str, t.Union[Security, Reference]]
    links: dict[str, t.Union[Link, Reference]]
    callbacks: dict[str, t.Union[Callback, Reference]]


@dataclasses.dataclass(frozen=True)
class OpenAPI:
    openapi: str
    info: Info
    paths: Paths
    components: Components
    servers: t.Optional[list[Server]] = None
    security: t.Optional[list[Security]] = None
    tags: t.Optional[list[Tag]] = None
    externalDocs: t.Optional[ExternalDocs] = None


class OpenAPISpec:
    OPENAPI_VERSION = "3.1.0"

    def __init__(
        self,
        title: str,
        version: str,
        description: t.Optional[str] = None,
        terms_of_service: t.Optional[str] = None,
        contact: t.Optional[Contact] = None,
        license: t.Optional[License] = None,
    ):
        self.spec = OpenAPI(
            openapi=self.OPENAPI_VERSION,
            info=Info(
                title=title,
                version=version,
                description=description,
                termsOfService=terms_of_service,
                contact=contact,
                license=license,
            ),
            paths=Paths({}),
            components=Components(
                schemas={},
                responses={},
                parameters={},
                examples={},
                requestBodies={},
                headers={},
                securitySchemes={},
                links={},
                callbacks={},
            ),
        )

    def add_path(self, path: str, item: Path):
        self.spec.paths[path] = item

    def add_schema(self, name: str, item: t.Union[Schema, Reference]):
        self.spec.components.schemas[name] = item

    def add_response(self, name: str, item: t.Union[Response, Reference]):
        self.spec.components.responses[name] = item

    def add_parameter(self, name: str, item: t.Union[Parameter, Reference]):
        self.spec.components.parameters[name] = item

    def add_example(self, name: str, item: t.Union[Example, Reference]):
        self.spec.components.examples[name] = item

    def add_request_body(self, name: str, item: t.Union[RequestBody, Reference]):
        self.spec.components.requestBodies[name] = item

    def add_header(self, name: str, item: t.Union[Header, Reference]):
        self.spec.components.headers[name] = item

    def add_security(self, name: str, item: t.Union[Security, Reference]):
        self.spec.components.securitySchemes[name] = item

    def add_link(self, name: str, item: t.Union[Link, Reference]):
        self.spec.components.links[name] = item

    def add_callback(self, name: str, item: t.Union[Callback, Reference]):
        self.spec.components.callbacks[name] = item

    def asdict(self, obj=None) -> t.Any:
        if obj is None:
            return self.asdict(dataclasses.asdict(self.spec))

        if isinstance(obj, list):
            return [self.asdict(i) for i in obj]

        if isinstance(obj, dict):
            return {{"ref": "$ref", "in_": "in"}.get(k, k): self.asdict(v) for k, v in obj.items() if v is not None}

        if dataclasses.is_dataclass(obj) and not isinstance(obj, type):
            return self.asdict(dataclasses.asdict(obj))

        return obj



import importlib
import os
import typing as t

from flama.injection import Component
from flama.models.base import Model
from flama.serialize import load
from flama.serialize.types import Framework

__all__ = ["ModelComponent", "ModelComponentBuilder"]


class ModelComponent(Component):
    def __init__(self, model):
        self.model = model

    def get_model_type(self) -> type[Model]:
        return self.model.__class__  # type: ignore[no-any-return]


class ModelComponentBuilder:
    @classmethod
    def _get_model_class(cls, framework: Framework) -> type[Model]:
        try:
            module, class_name = {
                Framework.torch: ("pytorch", "PyTorchModel"),
                Framework.sklearn: ("sklearn", "SKLearnModel"),
                Framework.tensorflow: ("tensorflow", "TensorFlowModel"),
                Framework.keras: ("tensorflow", "TensorFlowModel"),
            }[framework]
        except KeyError:  # pragma: no cover
            raise ValueError("Wrong framework")

        model_class: type[Model] = getattr(importlib.import_module(f"flama.models.models.{module}"), class_name)
        return model_class

    @classmethod
    def load(cls, path: t.Union[str, os.PathLike]) -> ModelComponent:
        load_model = load(path)
        parent = cls._get_model_class(load_model.meta.framework.lib)
        model_class = type(parent.__name__, (parent,), {})
        model_obj = model_class(load_model.model, load_model.meta, load_model.artifacts)

        class SpecificModelComponent(ModelComponent):
            def resolve(self) -> model_class:  # type: ignore[valid-type]
                return self.model  # type: ignore[no-any-return]

        return SpecificModelComponent(model_obj)
